

import numpy as np
import datastock as ds


import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from mpl_toolkits.mplot3d import Axes3D


from .. import version


_GITHUB = "https://github.com/ToFuProject/tofu"
_WINDEF = f"tofu {version.__version__} - report issues at {_GITHUB}"


_DFS = {
    1: (8, 5),
    2: (16, 10),
    3: (16, 10),
    4: (16, 10),
}


# #################################################################
# #################################################################
#               proj
# #################################################################


def _proj(
    proj=None,
    pall=None,
):

    if pall is None:
        pall = ['cross', 'hor', '3d', 'camera']

    # proj
    if proj == 'all':
        proj = pall

    if isinstance(proj, str):
        proj = [proj]

    proj = ds._generic_check._check_var_iter(
        proj, 'proj',
        default=pall,
        allowed=pall,
        types=list,
        types_iter=str,
    )
    return proj


# #################################################################
# #################################################################
#               Default axes
# #################################################################


# Generic
def get_dax_diag(
    proj=None,
    dmargin=None,
    fs=None,
    tit=None,
    wintit=None,
    is2d=None,
    key_cam=None,
):

    # ----------------------
    # Check inputs

    # fs
    fs = ds._generic_check._check_var(
        fs, 'fs',
        types=tuple,
        default=_DFS[len(proj)],
    )

    # tit
    tit = ds._generic_check._check_var(
        tit, 'tit',
        types=str,
        default='',
    )

    # wintit
    wintit = ds._generic_check._check_var(
        wintit, 'wintit',
        types=str,
        default=_WINDEF,
    )

    # is2d
    is2d = ds._generic_check._check_var(
        is2d, 'is2d',
        types=bool,
        default=False,
    )

    # -------------
    # Create figure

    fig = plt.figure(figsize=fs)
    fig.canvas.set_window_title(wintit)
    fig.suptitle(tit, size=12, fontweight='bold')

    # ------------
    # Populate dax

    dax = {}

    # Populate with default axes if necessary
    if len(proj) == 1:
        gs, dgs = _ax_single(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
            key_cam=key_cam,
        )

    elif len(proj) == 2:
        gs, dgs = _ax_double(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
            key_cam=key_cam,
        )

    elif len(proj) == 3:
        gs, dgs = _ax_3(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
            key_cam=key_cam,
        )
    else:
        gs, dgs = _ax_4(
            fig=fig,
            dmargin=dmargin,
            proj=proj,
            key_cam=key_cam,
        )


    # ----------------------
    # create figure and axes

    dax = {}
    for kk, vv in dgs.items():
        if vv['proj'] == '3d':
            ax = fig.add_subplot(
                gs[vv['ind']],
                projection='3d',
            )
        else:
            ax = fig.add_subplot(gs[vv['ind']])

        _ax_set(ax=ax, proj=vv['proj'], is2d=is2d)
        dax[kk] = ax

    return dax


# #################################################################
# #################################################################
#               Default axes
# #################################################################


def _ax_single(
    fig=None,
    dmargin=None,
    proj=None,
    key_cam=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.12, 'top': 0.9,
            'left': 0.10, 'right': 0.95,
        },
    )

    # ----------------------
    # prepare
    
    nrows = 1
    nc = 1 if key_cam is None else len(key_cam)
    if 'camera' in proj:
        nrows *= nc
        
    gs = gridspec.GridSpec(ncols=1, nrows=nrows, **dmargin)
        
    dgs = {}
    for ii, pp in enumerate(proj):
        if pp == 'camera':
            for jj, k0 in enumerate(key_cam):
                ind = (ii, 0)
                dgs[k0] = {'proj': 'camera', 'ind': ind}
        else:
            ind = (slice(0, nc), 0)
            dgs[pp] = {'proj': pp, 'ind': ind}

    return gs, dgs


def _ax_double(
    fig=None,
    dmargin=None,
    proj=None,
    key_cam=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.12, 'top': 0.9,
            'left': 0.08, 'right': 0.98,
            'wspace': 0.20, 'hspace': 0.05,
        },
    )

    # ----------------------
    # prepare
    
    nrows = 1
    nc = 1 if key_cam is None else len(key_cam)
    if 'camera' in proj:
        nrows *= nc
        
    gs = gridspec.GridSpec(ncols=2, nrows=nrows, **dmargin)
        
    dgs = {}
    for ii, pp in enumerate(proj):
        if pp == 'camera':
            for jj, k0 in enumerate(key_cam):
                ind = (jj, ii)
                dgs[k0] = {'proj': 'camera', 'ind': ind}
        else:
            ind = (slice(0, nc), ii)
            dgs[pp] = {'proj': pp, 'ind': ind}

    return gs, dgs


def _ax_3(
    fig=None,
    dmargin=None,
    proj=None,
    key_cam=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.12, 'top': 0.9,
            'left': 0.06, 'right': 0.98,
            'wspace': 0.10, 'hspace': 0.05,
        },
    )

    # ----------------------
    # prepare
    
    nrows = 2
    nc = 1 if key_cam is None else len(key_cam)
    if 'camera' in proj:
        nrows *= nc
        
    gs = gridspec.GridSpec(ncols=2, nrows=nrows, **dmargin)
        
    dgs = {}
    for ii, pp in enumerate(proj):
        if pp == 'camera':
            for jj, k0 in enumerate(key_cam):
                ind = (slice(2*jj, 2*(jj + 1)), 1)
                dgs[k0] = {'proj': 'camera', 'ind': ind}
        else:
            ind = (slice(ii*nc, nc*(ii+1)), ii // 2)
            dgs[pp] = {'proj': pp, 'ind': ind}

    return gs, dgs


def _ax_4(
    fig=None,
    dmargin=None,
    proj=None,
    key_cam=None,
):

    # ------------
    # check inputs

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.08, 'top': 0.90,
            'left': 0.06, 'right': 0.96,
            'wspace': 0.20, 'hspace': 0.20,
            # 'width_ratios': [0.6, 0.4],
            # 'height_ratios': [0.4, 0.6],
        },
    )

    # ----------------------
    # prepare
    
    nrows = 2
    nc = 1 if key_cam is None else len(key_cam)
    if 'camera' in proj:
        nrows *= nc

    gs = gridspec.GridSpec(ncols=2, nrows=nrows, **dmargin)
    
    dgs = {}
    for ii, pp in enumerate(proj):
        if pp == 'camera':
            for jj, k0 in enumerate(key_cam):
                ind = (nc + jj, 1)
                dgs[k0] = {'proj': 'camera', 'ind': ind}
        else:
            ind = (slice((ii % 2)*nc, (ii % 2)*nc + nc), ii//2)
            dgs[pp] = {'proj': pp, 'ind': ind}

    return gs, dgs


def _ax_set(ax=None, proj=None, is2d=None):

    if proj == 'cross':

        ax.set_xlabel('R (m)')
        ax.set_ylabel('Z (m)')
        ax.set_aspect('equal', adjustable='datalim')

    elif proj == 'hor':

        ax.set_xlabel('X (m)')
        ax.set_ylabel('Y (m)')
        ax.set_aspect('equal', adjustable='datalim')

    elif proj == '3d':

        ax.set_xlabel('X (m)')
        ax.set_ylabel('Y (m)')
        ax.set_zlabel('Z (m)')
        # ax.set_aspect('equal')

    elif proj == 'camera':

        if is2d:
            ax.set_xlabel('x0 (m)')
            ax.set_ylabel('x1 (m)')
            ax.set_aspect('equal', adjustable='datalim')
        else:
            ax.set_xlabel('ind')

    return



# -*- coding: utf-8 -*-
#! /usr/bin/python


"""
The openadas-compatibility module of tofu

"""


import traceback


try:
    try:
        from tofu.tomotok2tofu._core import *
    except Exception:
        from ._core import *
except Exception as err:
    msg = str(traceback.format_exc())
    msg += "\n\n    => optional sub-package tofu.tomotok2tofu not usable\n"
    raise Exception(msg)

del traceback



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import warnings
import shutil
import subprocess
from subprocess import PIPE
import itertools as itt

# Standard
import numpy as np
import matplotlib.pyplot as plt

# tofu-specific
from tofu import __version__
import tofu.openadas2tofu as tfoa

_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.openadas2tofu.test_03_core'
_TOFU_USER = os.path.join(os.path.expanduser("~"), '.tofu')
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(_here)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')


#######################################################
#
#     Setup and Teardown
#
#######################################################

def setup_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)
    # Recreating clean .tofu
    # out = subprocess.run(_CUSTOM, stdout=PIPE, stderr=PIPE)
    os.system('python '+_CUSTOM)


def teardown_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################


class Test01_openadas(object):

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_search_online(self):

        # Search by searchstr
        out = tfoa.step01_search_online(
            'ar+16 ADF15',
            verb=True,
            returnas=False,
        )
        assert out is None

        out = tfoa.step01_search_online(
            'ar+16 ADF11',
            verb=False,
            returnas=np.ndarray,
        )
        assert isinstance(out, np.ndarray)

        # Search by wavelength
        lret = [None, str, np.ndarray]
        lverb = [True, False]
        lelement = [None, 'ar', ['ar', 'w'], ('w',)]
        lcharge = [None, 14, [15, 16], (16,)]
        lres = ['transition', 'file']
        for comb in itt.product(lret, lverb, lelement, lcharge, lres):
            out = tfoa.step01_search_online_by_wavelengthA(
                lambmin=3.94, lambmax=4.,
                returnas=comb[0], verb=comb[1],
                element=comb[2], charge=comb[3],
                resolveby=comb[4],
            )
            assert out is comb[0] or isinstance(out, comb[0])

    def test02_download(self):
        out = tfoa.step02_download(
            filename='/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
            update=False,
            verb=True,
            returnas=False,
        )
        assert out is None

        out = tfoa.step02_download(
            filename='/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
            update=True,
            verb=False,
            returnas=str,
        )
        assert isinstance(out, str)

        out = tfoa.step02_download_all(
            lambmin=3.94,
            lambmax=4,
            element='Mo',
            update=False,
            verb=True,
        )
        assert out is None

        lf = [
            '/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
            '/adf11/scd74/scd74_ar.dat',
            '/adf15/pec40][w/pec40][w_ls][w64.dat',
            '/adf11/plt41/plt41_xe.dat',
        ]
        out = tfoa.step02_download_all(
            files=lf,
            update=False,
            verb=True,
        )
        assert out is None

        out = tfoa.step02_download_all(
            searchstr='ar+16 ADF15',
            include_partial=False,
            update=False,
            verb=True,
        )
        assert out is None

    def test03_readfiles(self):
        out = tfoa.step03_read(
            '/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
        )
        assert isinstance(out, dict)

        out = tfoa.step03_read(
            'adf11/plt41/plt41_xe.dat',
        )
        assert isinstance(out, dict)

        out = tfoa.step03_read_all(
            element='ar',
            charge=16,
            typ1='adf15',
            lambmin=3.94,
            lambmax=4.,
        )
        assert isinstance(out, dict)

        out = tfoa.step03_read_all(
            element='ar',
            charge=16,
            typ1='adf11',
            typ2='scd',
        )
        assert isinstance(out, dict)

        out = tfoa.step03_read_all(
            element='ar',
            charge=16,
            typ1='adf11',
            typ2='plt',
        )
        assert isinstance(out, dict)

    def test04_clear_downloads(self):
        tfoa.clear_downloads()





import warnings


import numpy as np
import scipy.linalg as scplin
import scipy.optimize as scpop
import scipy.sparse as scpsp


dfail = {}
try:
    import sksparse as sksp
except Exception as err:
    sksp = False
    dfail['sksparse'] = "For cholesk factorizations"

try:
    import scikits.umfpack as skumf
except Exception as err:
    skumf = False
    dfail['umfpack'] = "For faster sparse matrices"

if len(dfail) > 0:
    lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
    msg = (
        "Consider installing the following for faster inversions:\n"
        + "\n".join(lstr)
    )
    warnings.warn(msg)


# #############################################################################
# #############################################################################
#                      Basic routines - augmented tikhonov
# #############################################################################


def inv_linear_augTikho_dense(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    **kwdargs,
):
    """
    Linear algorithm for Phillips-Tikhonov regularisation
    Called "Augmented Tikhonov", dense matrix version
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    while niter < 2 or conv > conv_crit:

        # call solver
        sol = scplin.solve(
            TTn + mu0*R, Tyn,
            assume_a='pos',      # faster than 'sym'
            overwrite_a=True,    # no significant gain
            overwrite_b=False,   # True is faster, but a copy of Tyn is needed
            check_finite=False,  # small speed gain compared to True
            transposed=False,
        )  # 3

        # compute residu, regularity...
        res2 = np.sum((Tn.dot(sol)-yn)**2)  # residu**2
        reg = sol.dot(R.dot(sol))           # regularity term

        # update lamb, tau
        lamb = a0bis/(0.5*reg + b0)           # Update reg. param. estimate
        tau = a1bis/(0.5*res2 + b1)           # Update noise coef. estimate
        mu1 = (lamb/tau) * (2*a1bis/res2)**d  # rescale mu with noise estimate

        # Compute convergence variable
        if conv_reg:
            conv = np.abs(mu1 - mu0) / mu1
        else:
            sol2 = sol**2
            sol2max = np.max(sol2)
            sol2[sol2 < 0.001*sol2max] = 0.001*sol2max
            conv = np.sqrt(np.sum((sol - sol0)**2 / sol2) / nbs)

        # verb
        if verb >= 2:
            temp1 = f"{nchan} * {res2/nchan:.3e} + {mu1:.3e} * {reg:.3e}"
            temp2 = f"{res2 + mu1*reg:.3e}"
            temp = f"{temp1} = {temp2}"
            print(f"\t\t{niter} \t {temp}   {tau:.3e}   {conv:.3e}")

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_sparse(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    maxiter=None,
    tol=None,
    precond=None,       # test
    **kwdargs,
):
    """
    Linear algorithm for Phillips-Tikhonov regularisation
    Called "Augmented Tikhonov", sparese matrix version
    see InvLin_AugTikho_V1.__doc__ for details
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    while niter < 2 or conv > conv_crit:

        # sol = scpsp.linalg.spsolve(
        #    TTn + mu0*R, Tyn,
        #    permc_spec=None,
        #    use_umfpack=True,
        # )

        # seems faster
        sol, itconv = scpsp.linalg.cg(
            TTn + mu0*R, Tyn,
            x0=sol0,
            tol=tol,
            maxiter=maxiter,
            M=precond,
        )

        res2 = np.sum((Tn.dot(sol)-yn)**2)    # residu**2
        reg = sol.dot(R.dot(sol))             # regularity term

        lamb = a0bis/(0.5*reg + b0)           # Update reg. param. estimate
        tau = a1bis/(0.5*res2 + b1)           # Update noise coef. estimate
        mu1 = (lamb/tau) * (2*a1bis/res2)**d  # rescale mu with noise estimate

        # Compute convergence variable
        if conv_reg:
            conv = np.abs(mu1 - mu0) / mu1
        else:
            sol2 = sol**2
            sol2max = np.max(sol2)
            sol2[sol2 < 0.001*sol2max] = 0.001*sol2max
            conv = np.sqrt(np.sum((sol - sol0)**2 / sol2) / nbs)

        # verb
        if verb >= 2:
            temp1 = f"{nchan} * {res2/nchan:.3e} + {mu1:.3e} * {reg:.3e}"
            temp2 = f"{res2 + mu1*reg:.3e}"
            temp = f"{temp1} = {temp2}"
            print(
                f"\t\t{niter} \t {temp}   {tau:.3e}   {conv:.3e}"
            )

        sol0[:] = sol[:]            # Update reference solution
        niter += 1                  # Update number of iterations
        mu0 = mu1
    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_chol_dense(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    **kwdargs,
):
    """
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    while niter < 2 or conv > conv_crit:
        try:
            # choleski decomposition requires det(TT + mu0*LL) != 0
            # (chol(A).T * chol(A) = A
            chol = scplin.cholesky(
                TTn + mu0*R,
                lower=False,
                check_finite=False,
                overwrite_a=False,
            )
            # Use np.linalg.lstsq for double-solving the equation
            sol = scplin.cho_solve(
                (chol, False), Tyn,
                overwrite_b=None,
                check_finite=True,
            )
        except Exception as err:
            # call solver
            sol = scplin.solve(
                TTn + mu0*R, Tyn,
                assume_a='sym',         # chol failed => not 'pos'
                overwrite_a=True,       # no significant gain
                overwrite_b=False,      # True faster, but a copy of Tyn needed
                check_finite=False,     # small speed gain compared to True
                transposed=False,
            )  # 3

        # compute residu, regularity...
        res2 = np.sum((Tn.dot(sol)-yn)**2)  # residu**2
        reg = sol.dot(R.dot(sol))           # regularity term

        # update lamb, tau
        lamb = a0bis/(0.5*reg + b0)           # Update reg. param. estimate
        tau = a1bis/(0.5*res2 + b1)           # Update noise coef. estimate
        mu1 = (lamb/tau) * (2*a1bis/res2)**d  # mu rescale with noise estimate

        # Compute convergence variable
        if conv_reg:
            conv = np.abs(mu1 - mu0) / mu1
        else:
            sol2 = sol**2
            sol2max = np.max(sol2)
            sol2[sol2 < 0.001*sol2max] = 0.001*sol2max
            conv = np.sqrt(np.sum((sol - sol0)**2 / sol2) / nbs)

        # verb
        if verb >= 2:
            temp1 = f"{nchan} * {res2/nchan:.3e} + {mu1:.3e} * {reg:.3e}"
            temp2 = f"{res2 + mu1*reg:.3e}"
            temp = f"{temp1} = {temp2}"
            print(f"\t\t{niter} \t {temp}   {tau:.3e}   {conv:.3e}")

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_chol_sparse(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    **kwdargs,
):
    """
    Linear algorithm for Phillips-Tikhonov regularisation
    Called "Augmented Tikhonov"

    Augmented in the sense that bayesian statistics are combined
        with standard Tikhonov regularisation
    Determines both noise (common multiplicative coefficient) and
        regularisation parameter automatically
    We assume here that all arrays are scaled (noise, conditioning...)
    Sparse matrixes are also prefered to speed-up the computation

    In this method:
      tau is an approximation of the inverse of the noise coefficient
      lamb is an approximation of the regularisation parameter

    N.B.: The noise and reg. param. have probability densities of the form:
        f(x) = x^(a-1) * exp(-bx)
    This function's maximum is in x = (a-1)/b, so a = b+1 gives a maximum at 1.
        (a0, b0) for the reg. param.
        (a1, b1) for the noise estimate

    Ref:
      [1] Jin B., Zou J., Inverse Problems, vol.25, nb.2, 025001, 2009
      [2] http://www.math.uni-bremen.de/zetem/cms/media.php/250/nov14talk_jin%20bangti.pdf
      [3] Kazufumi Ito, Bangti Jin, Jun Zou,
        "A New Choice Rule for Regularization Parameters in Tikhonov
        Regularization", Research report, University of Hong Kong, 2008
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    factor = None
    while niter < 2 or conv > conv_crit:
        try:
            # choleski decomposition requires det(TT + mu0*LL) != 0
            # A = (chol(A).T * chol(A)
            # optimal if matrix is csc
            if sksp is False:
                factor = scpsp.linalg.factorized(TTn + mu0*R)
                sol = factor(Tyn)
            else:
                if factor is None:
                    factor = sksp.cholmod.cholesky(
                        TTn + mu0*R,
                        beta=0,
                        mode='auto',
                        ordering_method='default',
                        use_long=False,
                    )
                else:
                    # re-use same factor
                    factor.cholesky_inplace(TTn + mu0*R, beta=0)
                sol = factor.solve_A(Tyn)
        except Exception as err:
            # call solver
            sol = scpsp.linalg.spsolve(
                TTn + mu0*R, Tyn,
                permc_spec=None,
                use_umfpack=True,
            )

        # compute residu, regularity...
        res2 = np.sum((Tn.dot(sol)-yn)**2)  # residu**2
        reg = sol.dot(R.dot(sol))           # regularity term

        # update lamb, tau
        lamb = a0bis/(0.5*reg + b0)             # Update reg. param. estimate
        tau = a1bis/(0.5*res2 + b1)             # Update noise coef. estimate
        mu1 = (lamb/tau) * (2*a1bis/res2)**d    # Update reg. param. rescaling

        # Compute convergence variable
        if conv_reg:
            conv = np.abs(mu1 - mu0) / mu1
        else:
            sol2 = sol**2
            sol2max = np.max(sol2)
            sol2[sol2 < 0.001*sol2max] = 0.001*sol2max
            conv = np.sqrt(np.sum((sol - sol0)**2 / sol2) / nbs)

        # verb
        if verb >= 2:
            temp1 = f"{nchan} * {res2/nchan:.3e} + {mu1:.3e} * {reg:.3e}"
            temp2 = f"{res2 + mu1*reg:.3e}"
            temp = f"{temp1} = {temp2}"
            print(f"\t\t{niter} \t {temp}   {tau:.3e}   {conv:.3e}")

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_pos_dense(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    # specific
    method=None,
    options=None,
    bounds=None,
    func_val=None,
    func_jac=None,
    func_hess=None,
    **kwdargs,
):
    """
    Quadratic algorithm for Phillips-Tikhonov regularisation
    Alternative to the linear version with positivity constraint
    see TFI.InvLin_AugTikho_V1.__doc__ for details
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    while niter < 2 or conv > conv_crit:
        # quadratic method for positivity constraint
        sol = scpop.minimize(
            func_val, sol0,
            args=(mu0, Tn, yn, TTn, Tyn),
            jac=func_jac,
            hess=func_hess,
            method=method,
            bounds=bounds,
            options=options,
        ).x

        # compute residu, regularity...
        res2 = np.sum((Tn.dot(sol)-yn)**2)  # residu**2
        reg = sol.dot(R.dot(sol))           # regularity term

        # update lamb, tau
        lamb = a0bis/(0.5*reg + b0)             # Update reg. param. estimate
        tau = a1bis/(0.5*res2 + b1)             # Update noise coef. estimate
        mu1 = (lamb/tau) * (2*a1bis/res2)**d    # Update reg. param. rescaling

        # Compute convergence variable
        if conv_reg:
            conv = np.abs(mu1 - mu0) / mu1
        else:
            sol2 = sol**2
            sol2max = np.max(sol2)
            sol2[sol2 < 0.001*sol2max] = 0.001*sol2max
            conv = np.sqrt(np.sum((sol - sol0)**2 / sol2) / nbs)

        # verb
        if verb >= 2:
            temp1 = f"{nchan} * {res2/nchan:.3e} + {mu1:.3e} * {reg:.3e}"
            temp2 = f"{res2 + mu1*reg:.3e}"
            temp = f"{temp1} = {temp2}"
            print(f"\t\t{niter} \t {temp}   {tau:.3e}   {conv:.3e}")

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


# #############################################################################
# #############################################################################
#               Basic routines - discrepancy principle
# #############################################################################


def inv_linear_DisPrinc_sparse(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    mu0=None,
    precond=None,
    verb=None,
    verb2head=None,
    # specific
    chi2n_tol=None,
    chi2n_obj=None,
    maxiter=None,
    tol=None,
    **kwdargs,
):
    """
    Discrepancy principle: find mu such that chi2n = 1 +/- tol
    """

    niter = 0
    lchi2n = np.array([np.sum((Tn.dot(sol0) - yn)**2) / nchan])
    lmu = np.array([mu0])
    chi2n_obj_log = np.log(chi2n_obj)

    # verb
    if verb >= 2:
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {lchi2n[0]:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*lchi2n[0] + mu0*reg:.3e}",
            end='\n',
        )

    while niter == 0 or np.abs(lchi2n[-1] - chi2n_obj) > chi2n_tol:
        sol, itconv = scpsp.linalg.cg(
            TTn + lmu[-1]*R, Tyn,
            x0=sol0,
            tol=tol,
            maxiter=maxiter,
            M=precond,
        )

        lchi2n = np.append(lchi2n, np.sum((Tn.dot(sol) - yn)**2) / nchan)

        if niter == 0:
            if lchi2n[-1] >= chi2n_obj + chi2n_tol:
                lmu = np.append(lmu, lmu[-1] / 50.)
            elif lchi2n[-1] <= chi2n_obj - chi2n_tol:
                lmu = np.append(lmu, lmu[-1] * 50.)
            else:
                lmu = np.append(lmu, lmu[-1])
        elif niter == 1 or (
            np.all(lchi2n >= chi2n_obj + chi2n_tol)
            or np.all(lchi2n <= chi2n_obj - chi2n_tol)
        ):
            if lchi2n[-1] >= chi2n_obj + chi2n_tol:
                lmu = np.append(lmu, lmu[-1] / 50.)
            else:
                lmu = np.append(lmu, lmu[-1] * 50.)
        else:
            if lmu[-2] == lmu[-1]:
                # if the algo is stuck => break to avoid infinite loop
                ind = np.argmin(lchi2n[1:] - chi2n_obj)
                lmu[-1] = lmu[ind]
                lchi2n[-1] = lchi2n[ind]
                sol, itconv = scpsp.linalg.cg(
                    TTn + lmu[-1]*R, Tyn,
                    x0=sol0,
                    tol=tol,
                    maxiter=maxiter,
                    M=precond,
                )
                break
            else:
                indsort = np.argsort(lchi2n[1:])
                lmu = np.append(lmu, np.exp(np.interp(
                    chi2n_obj_log,
                    np.log(lchi2n[1:])[indsort],
                    np.log(lmu)[indsort]
                )))

        # verb
        if verb >= 2:
            reg = sol.dot(R.dot(sol))
            res2 = np.sum((Tn.dot(sol)-yn)**2)
            temp1 = f"{nchan} * {lchi2n[-1]:.3e} + {lmu[-1]:.3e} * {reg:.3e}"
            temp2 = f"{res2 + lmu[-1]*reg:.3e}"
            temp = f"{temp1} = {temp2}"
            print(f"\t\t{niter} \t {temp}")
            print(f"\t\t{niter} \t {temp}   \t   {np.abs(lchi2n[-1] - chi2n_obj):.3e}")

        sol0[:] = sol
        niter += 1

    reg = sol.dot(R.dot(sol))           # regularity term
    return sol, lmu[-1], lchi2n[-1], reg, niter, None


# #############################################################################
# #############################################################################
#               Non-regularized - least squares
# #############################################################################


def inv_linear_leastsquares_bounds(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    mu0=None,       # useless
    precond=None,
    verb=None,
    verb2head=None,
    # specific
    A=None,
    b=None,
    maxiter=None,
    tol=None,
    lsmr_tol=None,
    lsq_solver=None,
    method=None,
    bounds=None,
    dconstraints=None,
    **kwdargs,
):
    """
    Discrepancy principle: find mu such that chi2n = 1 +/- tol
    """

    # ---------
    # check inputs

    verbscp = verb
    if method is None:
        method = 'trf'
    if lsq_solver is None:
        if scpsp.issparse(Tn):
            lsq_solver = 'lsmr'
        else:
            lsq_solver = 'exact'

    # ------------
    # optimization

    # verb
    if verb >= 2:
        chi2n0 = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        print(
            f"{verb2head}\n\t\t\t {nchan} * {chi2n0:.3e} = {nchan*chi2n0:.3e}",
            end='\n',
        )

    # optimize
    res = scpop.lsq_linear(
        Tn,
        yn,
        bounds=bounds,
        method=method,
        tol=tol,
        lsq_solver=lsq_solver,
        lsmr_tol=lsmr_tol,
        max_iter=maxiter,
        verbose=verb,
    )

    # --------
    # return

    chi2n = np.sum((Tn.dot(res.x) - yn)**2) / nchan

    # verb
    if verb >= 2:
        temp = f"{nchan} * {chi2n:.3e} = {nchan * chi2n:.3e}"
        print(f"\t\t{res.nit} \t {temp}")

    if dconstraints is None:
        return res.x, None, chi2n, None, res.nit, None
    else:
        return (
            dconstraints['coefs'].dot(res.x) + dconstraints['offset']
        )





# Built-in
import itertools as itt

# Common
import numpy as np
from scipy.interpolate import BSpline
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.colors as mplcol
import matplotlib.gridspec as gridspec
from matplotlib.axes._axes import Axes
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.collections import PatchCollection
import matplotlib as mpl
from contourpy import contour_generator

# tofu
from tofu.version import __version__
from . import _def as _def

_GITHUB = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = 'tofu-%s        report issues / requests at %s'%(__version__, _GITHUB)

_QUIVERCOLOR = plt.cm.viridis(np.linspace(0, 1, 3))
_QUIVERCOLOR = np.array([[1., 0., 0., 1.],
                         [0., 1., 0., 1.],
                         [0., 0., 1., 1.]])
_QUIVERCOLOR = mplcol.ListedColormap(_QUIVERCOLOR)
_RAYS_NPTS = 10


# Generic
def _check_projdax_mpl(
    dax=None, proj=None,
    dmargin=None, fs=None, wintit=None,
):

    # ----------------------
    # Check inputs
    lproj = ['cross', 'hor', '3d', 'im']
    if proj is None:
        proj = 'all'
    if proj == 'all':
        proj = lproj
    if isinstance(proj, str):
        proj = [proj]
    c0 = isinstance(proj, list) and all([ss in lproj for ss in proj])
    if not c0:
        msg = (
            "Arg proj must be a list of allowed projections:\n"
            f"\t- allowed: {lproj}\n"
            f"\t- provided: {proj}"
        )
        raise Exception(msg)

    # ----------------------
    # Check dax
    lc = [
        dax is None,
        issubclass(dax.__class__, Axes),
        isinstance(dax, dict),
        isinstance(dax, list),
    ]
    assert any(lc)
    if lc[0]:
        dax = dict.fromkeys(proj)
    elif lc[1]:
        assert len(proj) == 1
        dax = {proj[0]: dax}
    elif lc[2]:
        lcax = [
            dax.get(pp) is None or issubclass(dax.get(pp).__class__, Axes)
            for pp in proj
        ]
        if not all(lcax):
            msg = (
                "Wrong key or axes in dax:\n"
                + "    - proj = {}".format(proj)
                + "    - dax = {}".format(dax)
            )
            raise Exception(msg)
    else:
        assert len(dax) == 2
        assert all(
            [ax is None or issubclass(ax.__class__, Axes) for ax in dax]
        )
        dax = {'cross': dax[0], 'hor': dax[1]}

    # Populate with default axes if necessary
    if proj == ['cross'] and dax.get('cross') is None:
        dax['cross'] = _def.Plot_LOSProj_DefAxes(
            'cross', fs=fs, dmargin=dmargin, wintit=wintit,
        )
    elif proj == ['hor'] and dax.get('hor') is None:
        dax['hor'] = _def.Plot_LOSProj_DefAxes(
            'hor', fs=fs, dmargin=dmargin, wintit=wintit,
        )
    elif proj == ['3d'] and dax.get('3d') is None:
        dax['3d'] = _def.Plot_3D_plt_Tor_DefAxes(
            fs=fs, dmargin=dmargin, wintit=wintit,
        )
    elif proj == ['im'] and dax.get('im') is None:
        dax['im'] = _def.Plot_CrystIm(
            fs=fs, dmargin=dmargin, wintit=wintit,
        )
    elif any([dax.get(k0) is None for k0 in proj]):
        dax = _def.Plot_AllCryst(
            fs=fs, dmargin=dmargin, wintit=wintit,
        )
    for kk in lproj:
        dax[kk] = dax.get(kk, None)
    return dax


# #################################################################
# #################################################################
#                   Generic geometry plot
# #################################################################
# #################################################################


def _CrystalBragg_plot_check(
    cryst=None, dcryst=None,
    det=None, ddet=None,
    res=None, element=None,
    color=None,
    pts_summit=None, pts1=None, pts2=None,
    xi=None, xj=None,
    rays_color=None, rays_npts=None,
    dleg=None, draw=True,
    miscut=None,
    wintit=None, tit=None,
):

    # plotting
    assert type(draw) is bool, "Arg draw must be a bool !"
    assert cryst is None or cryst.__class__.__name__ == 'CrystalBragg'
    if wintit is None:
        wintit = _WINTIT
    if tit is None:
        tit = False
    if dleg is None:
         dleg = _def.TorLegd

    # rays color
    if rays_color is None:
        if pts_summit is None:
            rays_color = 'k'
        else:
            rays_color = 'pts' if pts_summit.shape[2] > 1 else 'lamb'

    if rays_color in ['pts', 'lamb']:
        pass
    else:
        try:
            rays_color = mplcol.to_rgba(rays_color)
        except Exception as err:
            msg = (
                "Arg rays_color must be either:\n"
                + "\t- 'pts': color by pts of origin\n"
                + "\t- 'lamb': color by wavelength\n"
                + "\t- a matplotlib color (e.g.: 'k', (0.1, 0.2, 0.1), ...)\n"
            )
            raise Exception(msg)

    # rays_npts
    if rays_npts is None:
        rays_npts = _RAYS_NPTS
    assert rays_npts >= 2, "Arg rays_npts must be a int >= 2"

    # elements
    lelement = ['s', 'c', 'r', 'o', 'v']
    if element is None:
        element = 'oscrv'
    c0 = (
        isinstance(element, str)
        and all([ss in lelement for ss in element.lower()])
    )
    if not c0:
        msg = ("Arg element must be str contain some of the following:\n"
               + "\t- 'o': outline\n"
               + "\t- 'c': center (of curvature sphere)\n"
               + "\t- 's': summit (geometrical center of crystal piece)\n"
               + "\t- 'r': rowland circle (along e1 direction)\n"
               + "\t- 'v': local unit vectors\n"
               + "You provided:\n{}".format(element))
        raise Exception(msg)
    element = element.lower()

    # cryst
    if element != '' and cryst is None:
        msg = (
            "cryst cannot be None if element contains any of:\n"
            + "\t- {}\n".format(lelement)
            + "You provided: {}".format(element)
        )
        raise Exception(msg)

    # vectors and outline
    if cryst is not None:
        nout, e1, e2, miscut = cryst.get_unit_vectors(
            miscut=miscut,
        )
        nin = -nout
        outline = cryst.sample_outline_plot(
            res=res,
            miscut=miscut,
        )

    # det
    if det is None:
        det = False
    if det is not False and any([ss in element for ss in 'ocv']):
        c0 = isinstance(det, dict) and 'cent' in det.keys()
        if c0 and 'o' in element:
            c0 = c0 and all(
                [ss in det.keys() for ss in ['outline', 'ei', 'ej']]
            )
        if c0 and 'v' in element:
            c0 = c0 and all([ss in det.keys() for ss in ['nout', 'ei', 'ej']])
        if not c0:
            msg = ("Arg det must be a dict with keys:\n"
                   + "\t- 'cent': center of the detector\n"
                   + "\t- 'nout': outward unit vector (normal to surface)\n"
                   + "\t- 'ei': first local coordinate unit vector\n"
                   + "\t- 'ej': second coordinate unit vector\n"
                   + "\t- 'outline': 2d local coordinates of outline\n"
                   + "\tAll (except outline) are 3d cartesian coordinates\n"
                   + "You provided:\n{}".format(det))
            raise Exception(msg)

    # pts
    lc = [pts_summit is not None, pts1 is not None, pts2 is not None]
    c0 = (
        (not any(lc))
        or all(lc)
    )
    if not c0:
        msg = (
            "pts_summit and pts1 and pts2 must be:\n"
            + "\t- all None\n"
            + "\t- all np.ndarray of same shape, with shape[0] == 3\n"
            + "  You provided:\n"
            + "\t- pts_summit: {}\n".format(pts_summit)
            + "\t- pts1: {}".format(pts1)
            + "\t- pts2: {}".format(pts2)
        )
        raise Exception(msg)
    if pts_summit is not None:
        if not (pts_summit.shape == pts1.shape == pts2.shape):
            msg = (
                "Args pts_summit, pts1 and pts2 must have the same shape!\n"
                + "  You provided:\n"
                + "\t- pts_summit.shape: {}\n".format(pts_summit.shape)
                + "\t- pts1.shape: {}\n".format(pts1.shape)
                + "\t- pts2.shape: {}\n".format(pts2.shape)
            )
            raise Exception(msg)

    if rays_color in ['pts', 'lamb'] and pts_summit is not None:
        if pts_summit.ndim not in [4, 5]:
            msg = (
                "For pts-wise or lambda-wise coloring, "
                + "input pts_summit must be 4d np.ndarray of shape "
                + "(3, nlamb, npts, ndtheta)\n"
                + "  You provided:\n"
                + "\t- pts_summit.shape = {}".format(pts_summit.shape)
            )
            raise Exception(msg)

    # rays
    rays = None
    if pts_summit is not None:
        # pts.shape = (3, nlamb, npts, ndtheta)
        # rays.shape = (3, nlamb, npts, ndtheta, 2, 2*rays_npts)
        shape = np.r_[pts1.shape, 1]
        k = np.linspace(0, 1, rays_npts)
        rays = np.concatenate((
            pts1[..., None] + k*(pts_summit-pts1)[..., None],
            pts_summit[..., None]
            + k[1:]*(pts2-pts_summit)[..., None],
            np.full(shape, np.nan),
        ), axis=-1)

        nlamb, npts, ndtheta, _, nk = rays.shape[1:]
        if rays_color in ['pts', 'lamb']:
            if rays_color == 'lamb':
                rays = rays.reshape(3, nlamb, npts*ndtheta*nk*2).swapaxes(1, 2)
            elif rays_color == 'pts':
                rays = rays.swapaxes(1, 2).reshape(
                    3, npts, nlamb*ndtheta*nk*2,
                ).swapaxes(1, 2)
        else:
            rays = rays.reshape(3, nlamb*npts*ndtheta*nk*2, order='C')

    # xi, xj
    lc = [xi is not None, xj is not None]
    c0 = (
        np.sum(lc) in [0, 2]
        and (
            not any(lc)
            or (lc[0] and xi.shape == xj.shape == pts1.shape[1:])
        )
    )
    if not c0:
        msg = (
            "Args xi, xj must be either both None of 2 array of same shape!\n"
            + "  Provided:\n\t- xi:\n{}\n\t- xj:\n{}".format(xi, xj)
        )
        raise Exception(msg)
    if lc[0]:
        if rays_color in ['pts', 'lamb']:
            if rays_color == 'lamb':
                xi = xi.reshape(nlamb, npts*ndtheta*2).T
                xj = xj.reshape(nlamb, npts*ndtheta*2).T
            elif rays_color == 'pts':
                xi = xi.swapaxes(0, 1).reshape(npts, nlamb*ndtheta*2).T
                xj = xj.swapaxes(0, 1).reshape(npts, nlamb*ndtheta*2).T
        else:
            xi = xi.ravel()
            xj = xj.ravel()

    # dict for plotting
    if color is None:
        color = False
    lkd = ['outline', 'cent', 'summit', 'rowland', 'vectors']
    # Avoid passing default by reference
    if dcryst is None:
        dcryst = dict({k0: dict(v0)
                       for k0, v0 in _def._CRYSTAL_PLOT_DDICT.items()})
    else:
        dcryst = dict({k0: dict(v0) for k0, v0 in dcryst.items()})

    for k0 in lkd:
        if dcryst.get(k0) is None:
            dcryst[k0] = dict(_def._CRYSTAL_PLOT_DDICT[k0])
        if dcryst[k0].get('color') is None:
            if cryst is not None and cryst._dmisc.get('color') is not None:
                dcryst[k0]['color'] = cryst._dmisc['color']
        if color is not False:
            dcryst[k0]['color'] = color
    if ddet is None:
        # Avoid passing default by reference
        ddet = dict({k0: dict(v0)
                     for k0, v0 in _def._DET_PLOT_DDICT.items()})
    else:
        ddet = dict({k0: dict(v0) for k0, v0 in ddet.items()})
    for k0 in lkd:
        if ddet.get(k0) is None:
            ddet[k0] = dict(_def._DET_PLOT_DDICT[k0])
        if color is not False:
            ddet[k0]['color'] = color

    return (
        dcryst, det, ddet,
        nout, nin, e1, e2, xi, xj, outline, element, color,
        rays, rays_color, rays_npts,
        dleg, wintit,
    )


def CrystalBragg_plot(
    cryst=None, dcryst=None,
    det=None, ddet=None,
    dax=None, proj=None, res=None, element=None,
    color=None,
    pts_summit=None, pts1=None, pts2=None,
    xi=None, xj=None,
    rays_color=None, rays_npts=None,
    dleg=None, draw=True, fs=None, dmargin=None,
    miscut=None,
    wintit=None, tit=None,
):

    # ---------------------
    # Check / format inputs

    (
        dcryst, det, ddet,
        nout, nin, e1, e2, xi, xj, outline, element, color,
        rays, rays_color, rays_npts,
        dleg, wintit,
    ) = _CrystalBragg_plot_check(
        cryst=cryst, dcryst=dcryst,
        det=det, ddet=ddet,
        res=res, element=element,
        color=color,
        pts_summit=pts_summit, pts1=pts1, pts2=pts2,
        xi=xi, xj=xj,
        rays_color=rays_color, rays_npts=rays_npts,
        dleg=dleg, draw=draw,
        miscut=miscut,
        wintit=wintit, tit=tit,
    )

    # ---------------------
    # call plotting functions
    dax = _CrystalBragg_plot(
        cryst=cryst, dcryst=dcryst,
        det=det, ddet=ddet,
        nout=nout, nin=nin, e1=e1, e2=e2, outline=outline,
        proj=proj, dax=dax, element=element,
        rays=rays, rays_color=rays_color, rays_npts=rays_npts,
        xi=xi, xj=xj,
        draw=draw, dmargin=dmargin, fs=fs, wintit=wintit,
    )

    # recompute the ax.dataLim
    ax0 = None
    for kk, vv in dax.items():
        if vv is None:
            continue
        dax[kk].relim()
        dax[kk].autoscale_view()
        if dleg is not False:
            dax[kk].legend(**dleg)
        ax0 = vv

    # set title
    if tit != False:
        ax0.figure.suptitle(tit)
    if draw:
        ax0.figure.canvas.draw()
    return dax


def _CrystalBragg_plot(
    cryst=None, dcryst=None,
    det=None, ddet=None,
    nout=None, nin=None, e1=None, e2=None,
    outline=None,
    proj=None, dax=None, element=None,
    rays=None, rays_color=None, rays_npts=None,
    xi=None, xj=None,
    quiver_cmap=None, draw=True,
    dmargin=None, fs=None, wintit=None,
):

    # ---------------------
    # Check / format inputs

    if 'v' in element and quiver_cmap is None:
        quiver_cmap = _QUIVERCOLOR

    # ---------------------
    # Prepare axe and data

    dax = _check_projdax_mpl(
        dax=dax, proj=proj,
        dmargin=dmargin, fs=fs, wintit=wintit,
    )

    if 's' in element or 'v' in element:
        summ = cryst._dgeom['summit']
    if 'c' in element:
        cent = cryst._dgeom['center']
    if 'r' in element:
        ang = np.linspace(0, 2.*np.pi, 200)
        rr = 0.5*cryst._dgeom['rcurve']
        row = cryst._dgeom['summit'] + rr*nin
        row = (row[:, None]
               + rr*(np.cos(ang)[None, :]*nin[:, None]
                     + np.sin(ang)[None, :]*e1[:, None]))

    # ---------------------
    # plot
    cross = dax.get('cross') is not None
    hor = dax.get('hor') is not None
    d3 = dax.get('3d') is not None
    im = dax.get('im') is not None and xi is not None

    if 'o' in element:
        if cross:
            dax['cross'].plot(
                np.hypot(outline[0, :], outline[1, :]),
                outline[2, :],
                label=cryst.Id.NameLTX+' outline',
                **dcryst['outline'],
            )
        if hor:
            dax['hor'].plot(
                outline[0, :], outline[1, :],
                label=cryst.Id.NameLTX+' outline',
                **dcryst['outline'],
            )
        if d3:
            dax['3d'].plot(
                outline[0, :], outline[1, :], outline[2, :],
                label=cryst.Id.NameLTX+' outline',
                **dcryst['outline'],
            )

    if 's' in element:
        if cross:
            dax['cross'].plot(
                np.hypot(summ[0], summ[1]), summ[2],
                label=cryst.Id.NameLTX+" summit",
                **dcryst['summit'],
            )
        if hor:
            dax['hor'].plot(
                summ[0], summ[1],
                label=cryst.Id.NameLTX+" summit",
                **dcryst['summit'],
            )
        if d3:
            dax['3d'].plot(
                summ[0:1], summ[1:2], summ[2:3],
                label=cryst.Id.NameLTX+" summit",
                **dcryst['summit'],
            )

    if 'c' in element:
        if cross:
            dax['cross'].plot(
                np.hypot(cent[0], cent[1]), cent[2],
                label=cryst.Id.NameLTX+" center",
                **dcryst['cent'],
            )
        if hor:
            dax['hor'].plot(
                cent[0], cent[1],
                label=cryst.Id.NameLTX+" center",
                **dcryst['cent'],
            )
        if d3:
            dax['3d'].plot(
                cent[0:1], cent[1:2], cent[2:3],
                label=cryst.Id.NameLTX+" center",
                **dcryst['cent'],
            )

    if 'r' in element:
        if cross:
            dax['cross'].plot(
                np.hypot(row[0, :], row[1, :]), row[2, :],
                label=cryst.Id.NameLTX+' rowland',
                **dcryst['rowland'],
            )
        if hor:
            dax['hor'].plot(
                row[0, :], row[1, :],
                label=cryst.Id.NameLTX+' rowland',
                **dcryst['rowland'],
            )
        if d3:
            dax['3d'].plot(
                row[0, :], row[1, :], row[2, :],
                label=cryst.Id.NameLTX+' rowland',
                **dcryst['rowland'],
            )
    if 'v' in element:
        p0 = np.repeat(summ[:,None], 3, axis=1)
        v = np.concatenate((nout[:, None], e1[:, None], e2[:, None]), axis=1)
        if cross:
            pr = np.hypot(p0[0, :], p0[1, :])
            vr = np.hypot(p0[0, :]+v[0, :], p0[1, :]+v[1, :]) - pr
            dax['cross'].quiver(
                pr, p0[2, :],
                vr, v[2, :],
                np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                angles='xy', scale_units='xy',
                label=cryst.Id.NameLTX+" unit vect",
                **dcryst['vectors'],
            )
        if hor:
            dax['hor'].quiver(
                p0[0, :], p0[1, :],
                v[0, :], v[1, :],
                np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                angles='xy', scale_units='xy',
                label=cryst.Id.NameLTX+" unit vect",
                **dcryst['vectors'],
            )
        if d3:
            dax['3d'].quiver(
                p0[0, :], p0[1, :], p0[2, :],
                v[0, :], v[1, :], v[2, :],
                np.r_[0., 0.5, 1.],
                length=0.1,
                normalize=True,
                cmap=quiver_cmap,
                label=cryst.Id.NameLTX+" unit vect",
                **dcryst['vectors'],
            )

    # -------------
    # Detector
    if det is not False:
        if det.get('cent') is not None and 'c' in element:
            if cross:
                dax['cross'].plot(
                    np.hypot(det['cent'][0], det['cent'][1]),
                    det['cent'][2],
                    label="det_cent",
                    **ddet['cent'],
                )
            if hor:
                dax['hor'].plot(
                    det['cent'][0], det['cent'][1],
                    label="det_cent",
                    **ddet['cent'],
                )
            if d3:
                dax['3d'].plot(
                    det['cent'][0:1],
                    det['cent'][1:2],
                    det['cent'][2:3],
                    label="det_cent",
                    **ddet['cent'],
                )

        if det.get('nout') is not None and 'v' in element:
            assert det.get('ei') is not None and det.get('ej') is not None
            p0 = np.repeat(det['cent'][:, None], 3, axis=1)
            v = np.concatenate((det['nout'][:, None], det['ei'][:, None],
                                det['ej'][:, None]), axis=1)
            if cross:
                pr = np.hypot(p0[0, :], p0[1, :])
                vr = np.hypot(p0[0, :]+v[0, :], p0[1, :]+v[1, :]) - pr
                dax['cross'].quiver(
                    pr, p0[2, :],
                    vr, v[2, :],
                    np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                    angles='xy', scale_units='xy',
                    label="det unit vect",
                    **ddet['vectors'],
                )
            if hor:
                dax['hor'].quiver(
                    p0[0, :], p0[1, :],
                    v[0, :], v[1, :],
                    np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                    angles='xy', scale_units='xy',
                    label="det unit vect",
                    **ddet['vectors'],
                )
            if d3:
                dax['3d'].quiver(
                    p0[0, :], p0[1, :], p0[2, :],
                    v[0, :], v[1, :], v[2, :],
                    np.r_[0., 0.5, 1.],
                    length=0.1,
                    normalize=True,
                    cmap=quiver_cmap,
                    label="det unit vect",
                    **ddet['vectors'],
                )

        if det.get('outline') is not None and 'o' in element:
            det_out = (
                det['outline'][0:1, :]*det['ei'][:, None]
                + det['outline'][1:2, :]*det['ej'][:, None]
                + det['cent'][:, None]
            )

            if cross:
                dax['cross'].plot(
                    np.hypot(det_out[0, :], det_out[1, :]),
                    det_out[2, :],
                    label='det outline',
                    **ddet['outline'],
                )
            if hor:
                dax['hor'].plot(
                    det_out[0, :],
                    det_out[1, :],
                    label='det outline',
                    **ddet['outline'],
                )
            if d3:
                dax['3d'].plot(
                    det_out[0, :],
                    det_out[1, :],
                    det_out[2, :],
                    label='det outline',
                    **ddet['outline'],
                )
            if im:
                dax['im'].plot(
                    det['outline'][0, :],
                    det['outline'][1, :],
                    label='det outline',
                    **ddet['outline'],
                )

    # -------------
    # rays
    if rays is not None:
        if rays_color in ['pts', 'lamb']:
            if cross:
                dax['cross'].set_prop_cycle(None)
                dax['cross'].plot(
                    np.hypot(rays[0, :, :], rays[1, :, :]),
                    rays[2, :, :],
                    lw=1., ls='-',
                )
            if hor:
                dax['hor'].set_prop_cycle(None)
                dax['hor'].plot(
                    rays[0, :, :], rays[1, :, :],
                    lw=1., ls='-',
                )
            if d3:
                dax['3d'].set_prop_cycle(None)
                for ii in range(rays.shape[2]):
                    dax['3d'].plot(
                        rays[0, :, ii],
                        rays[1, :, ii],
                        rays[2, :, ii],
                        lw=1., ls='-',
                    )
            if im:
                dax['3d'].set_prop_cycle(None)
                dax['im'].plot(
                    xi, xj,
                    ls='None',
                    marker='.',
                    ms=6,
                )

        else:
            if cross:
                dax['cross'].set_prop_cycle(None)
                dax['cross'].plot(
                    np.hypot(rays[0, :], rays[1, :]),
                    rays[2, :],
                    color=rays_color, lw=1., ls='-',
                )
            if hor:
                dax['hor'].set_prop_cycle(None)
                dax['hor'].plot(
                    rays[0, :], rays[1, :],
                    color=rays_color, lw=1., ls='-',
                )
            if d3:
                dax['3d'].set_prop_cycle(None)
                dax['3d'].plot(
                    rays[0, :], rays[1, :], rays[2, :],
                    color=rays_color, lw=1., ls='-',
                )
            if im:
                dax['3d'].set_prop_cycle(None)
                dax['im'].plot(
                    xi, xj,
                    ls='None',
                    marker='.',
                    ms=6,
                )
    return dax


# #################################################################
# #################################################################
#                   Rocking curve plot
# #################################################################
# #################################################################


def CrystalBragg_plot_rockingcurve(
    func=None, bragg=None, lamb=None,
    sigma=None, npts=None,
    ang_units=None, axtit=None,
    color=None,
    legend=None, fs=None, ax=None,
):

    # Prepare
    if legend is None:
        legend = True
    if color is None:
        color = 'k'
    if ang_units is None:
        ang_units = 'deg'
    if axtit is None:
        axtit = 'Rocking curve'
    if sigma is None:
        sigma = 0.005*np.pi/180.
    if npts is None:
        npts = 1000
    angle = bragg + 3.*sigma*np.linspace(-1, 1, npts)
    curve = func(angle)
    lab = r"$\lambda = {:9.6} A$".format(lamb*1.e10)
    if ang_units == 'deg':
        angle = angle*180/np.pi
        bragg = bragg*180/np.pi

    # Plot
    if ax is None:
        if fs is None:
            fs = (8, 6)
        fig = plt.figure(figsize=fs)
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.set_title(axtit, size=12)
        ax.set_xlabel('angle ({})'.format(ang_units))
        ax.set_ylabel('reflectivity (adim.)')
    ax.plot(angle, curve, ls='-', lw=1., c=color, label=lab)
    ax.axvline(bragg, ls='--', lw=1, c=color)
    if legend is not False:
        ax.legend()
    return ax


# #################################################################
# #################################################################
#                   Bragg diffraction plot
# #################################################################
# #################################################################

# Deprecated ? re-use ?
def CrystalBragg_plot_approx_detector_params(Rrow, bragg, d, Z,
                                             frame_cent, nn):

    R = 2.*Rrow
    L = 2.*R
    ang = np.linspace(0., 2.*np.pi, 100)

    fig = plt.figure()
    ax = fig.add_axes([0.1,0.1,0.8,0.8], aspect='equal')

    ax.axvline(0, ls='--', c='k')
    ax.plot(Rrow*np.cos(ang), Rrow + Rrow*np.sin(ang), c='r')
    ax.plot(R*np.cos(ang), R + R*np.sin(ang), c='b')
    ax.plot(L*np.cos(bragg)*np.r_[-1,0,1],
            L*np.sin(bragg)*np.r_[1,0,1], c='k')
    ax.plot([0, d*np.cos(bragg)], [Rrow, d*np.sin(bragg)], c='r')
    ax.plot([0, d*np.cos(bragg)], [Z, d*np.sin(bragg)], 'g')
    ax.plot([0, L/10*nn[1]], [Z, Z+L/10*nn[2]], c='g')
    ax.plot(frame_cent[1]*np.cos(2*bragg-np.pi),
            Z + frame_cent[1]*np.sin(2*bragg-np.pi), c='k', marker='o', ms=10)

    ax.set_xlabel(r'y')
    ax.set_ylabel(r'z')
    ax.legend(loc='upper left', bbox_to_anchor=(1.05, 1.), frameon=False)
    return ax


def CrystalBragg_plot_xixj_from_braggangle(bragg=None, xi=None, xj=None,
                                           data=None, ax=None):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_axes([0.1,0.1,0.8,0.8], aspect='equal')

    for ii in range(len(bragg)):
        deg ='{0:07.3f}'.format(bragg[ii]*180/np.pi)
        ax.plot(xi[:,ii], xj[:,ii], '.', label='bragg %s'%deg)

    ax.set_xlabel(r'xi')
    ax.set_ylabel(r'yi')
    ax.legend(loc='upper left', bbox_to_anchor=(1.05, 1.), frameon=False)
    return ax


def CrystalBragg_plot_braggangle_from_xixj(xi=None, xj=None,
                                           bragg=None, angle=None,
                                           ax=None, plot=None,
                                           braggunits='rad', angunits='rad',
                                           leg=None, colorbar=None,
                                           fs=None, wintit=None,
                                           tit=None, **kwdargs):

    # Check inputs
    if isinstance(plot, bool):
        plot = 'contour'
    if fs is None:
        fs = (6, 6)
    if wintit is None:
        wintit = _WINTIT
    if tit is None:
        tit = False
    if colorbar is None:
        colorbar = True
    if leg is None:
        leg = False
    if leg is True:
        leg = {}

    # Prepare axes
    if ax is None:
        fig = plt.figure(figsize=fs)
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8],
                          aspect='equal', adjustable='box')
    dobj = {'phi': {'ax': ax}, 'bragg': {'ax': ax}}
    dobj['bragg']['kwdargs'] = dict(kwdargs)
    dobj['phi']['kwdargs'] = dict(kwdargs)
    dobj['phi']['kwdargs']['cmap'] = plt.cm.seismic

    # Clear cmap if colors provided
    if 'colors' in kwdargs.keys():
        if 'cmap' in dobj['bragg']['kwdargs'].keys():
            del dobj['bragg']['kwdargs']['cmap']
        if 'cmap' in dobj['phi']['kwdargs'].keys():
            del dobj['phi']['kwdargs']['cmap']

    # Plot
    if plot == 'contour':
        if 'levels' in kwdargs.keys():
            lvls = kwdargs['levels']
            del kwdargs['levels']
            obj0 = dobj['bragg']['ax'].contour(xi, xj, bragg, lvls,
                                               **dobj['bragg']['kwdargs'])
            obj1 = dobj['phi']['ax'].contour(xi, xj, angle, lvls,
                                             **dobj['phi']['kwdargs'])
        else:
            obj0 = dobj['bragg']['ax'].contour(xi, xj, bragg,
                                               **dobj['bragg']['kwdargs'])
            obj1 = dobj['phi']['ax'].contour(xi, xj, angle,
                                             **dobj['phi']['kwdargs'])
    elif plot == 'imshow':
        extent = (xi.min(), xi.max(), xj.min(), xj.max())
        obj0 = dobj['bragg']['ax'].imshow(bragg, extent=extent, aspect='equal',
                                          adjustable='datalim',
                                          **dobj['bragg']['kwdargs'])
        obj1 = dobj['phi']['ax'].imshow(angle, extent=extent, aspect='equal',
                                        adjustable='datalim',
                                        **dobj['phi']['kwdargs'])
    elif plot == 'pcolor':
        obj0 = dobj['bragg']['ax'].pcolor(xi, xj, bragg,
                                          **dobj['bragg']['kwdargs'])
        obj1 = dobj['phi']['ax'].pcolor(xi, xj, angle,
                                        **dobj['phi']['kwdargs'])
    dobj['bragg']['obj'] = obj0
    dobj['phi']['obj'] = obj1

    # Post polish
    for k0 in set(dobj.keys()):
        dobj[k0]['ax'].set_xlabel(r'xi (m)')
        dobj[k0]['ax'].set_ylabel(r'xj (m)')

    if colorbar is True:
        cax0 = plt.colorbar(dobj['bragg']['obj'], ax=dobj['bragg']['ax'])
        cax1 = plt.colorbar(dobj['phi']['obj'], ax=dobj['phi']['ax'])
        cax0.ax.set_title(r'$\theta_{bragg}$' + '\n' + r'($%s$)' % braggunits)
        cax1.ax.set_title(r'$ang$' + '\n' + r'($%s$)' % angunits)

    if leg is not False:
        ax.legend(**leg)
    if wintit is not False:
        ax.figure.canvas.manager.set_window_title(wintit)
    if tit is not False:
        ax.figure.suptitle(tit, size=10, weight='bold', ha='right')
    return ax


def CrystalBragg_plot_line_tracing_on_det(
    cryst=None, dcryst=None,
    lamb=None,
    xi=None, xj=None, xi_er=None, xj_er=None,
    power_ratio=None, dth=None, ndth=None, nn=None,
    xi_rc=None, xj_rc=None,
    xi_atprmax=None,
    bragg_atprmax=None,
    lamb_atprmax=None,
    det=None,
    johann=None, rocking=None,
    miscut=None,
    therm_exp=None,
    merge_rc_data=None,
    alpha0=None, temp0=None, TD=None, angles=None,
    id_temp0=None,
    ax=None, dleg=None, color=None,
    fs=None, dmargin=None, wintit=None, tit=None,
):

    # Check inputs
    # ------------

    if dleg is None:
        dleg = {
            'loc': 'upper left',
            'fontsize': 13,
        }
    if color is None:
        color = 'k'

    if fs is None:
        fs = (8, 8)
    if dmargin is None:
        dmargin = {'left': 0.15, 'right': 0.95,
                   'bottom': 0.08, 'top': 0.92,
                   'wspace': None, 'hspace': 0.4}

    if wintit is None:
        wintit = _WINTIT
    if tit is None:
        tit = "Ray-tracing on camera surface"
        if johann is True:
            tit += " - johann error"
        if rocking is True:
            tit += " - rocking curve"

    plot_err = johann is True or rocking is True
    markers = ['o', '^', 'D', 's', 'X']
    colors = ['r', 'g', 'c', 'b', 'k']

    # Plot
    # ------------

    if ax is None:
        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(1, 1, **dmargin)
        ax = fig.add_subplot(gs[0, 0], aspect='equal', adjustable='datalim')
        if wintit is not False:
            fig.canvas.manager.set_window_title(wintit)
        if tit is not False:
            fig.suptitle(tit, size=14, weight='bold')
        ax.set_xlabel(r'Pixel coordinate $x_{i}$ [m]', fontsize=15)
        ax.set_ylabel(r'Pixel coordinate $x_{j}$ [m]', fontsize=15)
        ax.set_xlim(
            det['outline'][0, :].min() - 0.01,
            det['outline'][0, :].max() + 0.01,
        )
        ax.set_ylim(
            det['outline'][1, :].min() - 0.01,
            det['outline'][1, :].max() + 0.01,
        )
    if det.get('outline') is not None:
        ax.plot(
            det['outline'][0, :], det['outline'][1, :],
            ls='-', lw=1., c='k',
        )
    aa = np.r_[cryst.dmat['alpha']]
    if therm_exp and merge_rc_data:
        bb = TD[id_temp0]
    elif therm_exp and not merge_rc_data:
        bb = temp0
    else:
        bb = 0.
    for ll in range(lamb.size):
        lab = (
            r'$\lambda$ = {} A'.format(np.round(lamb[ll]*1e10, 6)) + '\n'
            + r'$\Delta$T = {} °C, $\alpha$ = {} deg'.format(
                bb, aa[0]*(180./np.pi)
            )
        )
        l0, = ax.plot(
            xi[ll, :], xj[ll, :],
            ls='--', lw=1.,
            marker=markers[ll], ms=4.,
            c=color,
            label=lab,
        )
        if plot_err:
            ax.plot(
                xi_er[ll, ...], xj_er[ll, ...],
                ls='None', lw=1., c=l0.get_color(),
                ms=4, marker='.',
            )
    if merge_rc_data:
        for ll in range(lamb.size):
            for mm in range(ndth):
                if mm == int(ndth/2.):
                    label = r'At $x_j$=0.: $x_i$={}, $\lambda$={}A'.format(
                        np.round(xi_atprmax[ll], 6),
                        np.round(lamb_atprmax[ll], 16),
                        # np.round(bragg_atprmax[ll]*(180./np.pi), 4),
                    )
                else:
                    label = None
                pr1 = power_ratio[ll, 0, 0, 0, mm]
                pr2 = power_ratio[ll, 1, 0, 0, mm]
                ax.plot(
                    xi_rc[ll, mm, :], xj_rc[ll, mm, :],
                    ls='-', lw=1.,
                    c=l0.get_color(),
                    alpha=pr1 + pr2,
                    label=label,
                )

    if dleg is not False:
        ax.legend(**dleg)

    return ax


def CrystalBragg_plot_angular_shift_on_det_tracing(
    cryst=None, dcryst=None,
    lamb=None,
    din=None,
    na=None, nn=None,
    det=None,
    TD=None, angles=None,
    ax=None, dleg=None, color=None,
    fs=None, dmargin=None, wintit=None, tit=None,
):

    # Check inputs
    # ------------

    if dleg is None:
        dleg = {
            'loc': 'upper left',
            'fontsize': 13,
        }
    if color is None:
        color = 'k'
    if fs is None:
        fs = (12, 12)
    """
    if dmargin is None:
        dmargin = {'left': 0.15, 'right': 0.95,
                   'bottom': 0.08, 'top': 0.92,
                   'wspace': None, 'hspace': 0.4}
    """
    if wintit is None:
        wintit = _WINTIT
    if tit is None:
        tit = "Angular shift from the ideal line position"
    cmap = plt.cm.seismic  # viridis

    # Plot
    # ------------

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(1, 3)  # , **dmargin)
    ax0 = fig.add_subplot(gs[0, 0], aspect='equal', adjustable='datalim')
    ax0.set_title('Pixel offset [m]', fontsize=20)
    plt.xticks(fontsize=18)
    plt.yticks(fontsize=18)
    ax1 = fig.add_subplot(gs[0, 1], aspect='equal', adjustable='datalim')
    ax1.set_title(r'Spectral offset $[\AA]$', fontsize=20)
    plt.xticks(fontsize=18)
    plt.yticks(fontsize=18)
    ax2 = fig.add_subplot(gs[0, 2], aspect='equal', adjustable='datalim')
    ax2.set_title('Angular offset [mrad]', fontsize=20)
    plt.xticks(fontsize=18)
    plt.yticks(fontsize=18)
    if wintit is not False:
        fig.canvas.manager.set_window_title(wintit)
    if tit is not False:
        fig.suptitle(tit, size=14, weight='bold')
    ax0.set_ylabel(r'$\Delta$T ($T_{0}$=25°C)', fontsize=20)
    ax0.set_xlabel(r'$\alpha$ [mrad]', fontsize=20)
    ax1.set_xlabel(r'$\alpha$ [mrad]', fontsize=20)
    ax2.set_xlabel(r'$\alpha$ [mrad]', fontsize=20)

    extent = (angles.min()*1e3, angles.max()*1e3, TD.min(), TD.max())
    delta_xi = din['delta_xi'].reshape(
        din['delta_xi'].shape[0],
        din['delta_xi'].shape[1]
    )
    delta_lamb = din['delta_lamb'].reshape(
        din['delta_lamb'].shape[0],
        din['delta_lamb'].shape[1]
    )
    delta_bragg = din['delta_bragg'].reshape(
        din['delta_bragg'].shape[0],
        din['delta_bragg'].shape[1]
    )

    # Plot imshow maps
    cmap_xi = ax0.imshow(
        delta_xi,
        cmap=cmap,
        origin='lower',
        extent=extent,
        aspect='auto',
    )
    cbar0 = plt.colorbar(
        cmap_xi,
        orientation='vertical',
        ax=ax0,
    )
    cbar0.ax.tick_params(labelsize=18)
    cmap_lamb = ax1.imshow(
        delta_lamb,
        cmap=cmap,
        origin='lower',
        extent=extent,
        aspect='auto',
    )
    cbar1 = plt.colorbar(
        cmap_lamb,
        orientation='vertical',
        ax=ax1,
    )
    cbar1.ax.tick_params(labelsize=18)
    cmap_bragg = ax2.imshow(
        delta_bragg*1e3,
        cmap=cmap,
        origin='lower',
        extent=extent,
        aspect='auto',
    )
    cbar2 = plt.colorbar(
        cmap_bragg,
        orientation='vertical',
        ax=ax2,
    )
    cbar2.ax.tick_params(labelsize=18)


def CrystalBragg_plot_johannerror(
    xi, xj, lamb, phi, err_lamb, err_phi,
    err_lamb_units=None,
    err_phi_units=None,
    cmap=None, vmin=None, vmax=None,
    fs=None, dmargin=None, wintit=None, tit=None,
    angunits=None,
):

    # Check inputs
    # ------------

    if fs is None:
        fs = (14, 8)
    if cmap is None:
        cmap = plt.cm.viridis
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.99,
                   'bottom': 0.06, 'top': 0.92,
                   'wspace': None, 'hspace': 0.4}

    if angunits is None:
        angunits = 'rad'
    assert angunits in ['deg', 'rad']
    if angunits == 'deg':
        # bragg = bragg*180./np.pi
        phi = phi*180./np.pi
        err_phi = err_phi*180./np.pi
        err_phi_units = angunits

    if wintit is None:
        wintit = _WINTIT
    if tit is None:
        tit = False

    # pre-compute
    # ------------

    # extent
    extent = (xi.min(), xi.max(), xj.min(), xj.max())

    # Plot
    # ------------
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(1, 3, **dmargin)
    ax0 = fig.add_subplot(gs[0, 0], aspect='equal')     # adjustable='datalim')
    ax1 = fig.add_subplot(
        gs[0, 1], aspect='equal', sharex=ax0, sharey=ax0,
    )
    ax2 = fig.add_subplot(
        gs[0, 2], aspect='equal', sharex=ax0, sharey=ax0,
    )

    ax0.set_title('Iso-lamb and iso-phi at crystal summit')
    ax1.set_title(f'Focalization error on lamb ({err_lamb_units})')
    ax2.set_title(f'Focalization error on phi ({err_phi_units})')
    ax0.contour(xi, xj, lamb.T, 10, cmap=cmap)
    ax0.contour(xi, xj, phi.T, 10, cmap=cmap, linestyles='--')
    imlamb = ax1.imshow(
        err_lamb.T,
        extent=extent, aspect='equal',
        origin='lower', interpolation='nearest',
        vmin=vmin, vmax=vmax,
    )
    imphi = ax2.imshow(
        err_phi.T,
        extent=extent, aspect='equal',
        origin='lower', interpolation='nearest',
        vmin=vmin, vmax=vmax,
    )

    plt.colorbar(imlamb, ax=ax1)
    plt.colorbar(imphi, ax=ax2)

    if wintit is not False:
        fig.canvas.manager.set_window_title(wintit)
    if tit is not False:
        fig.suptitle(tit, size=14, weight='bold')

    return [ax0, ax1, ax2]


def CrystalBragg_plot_focal_error_summed(
    cryst=None, dcryst=None,
    lamb=None, bragg=None,
    error_lambda=None,
    ddist=None, di=None,
    ddist0=None, di0=None, dj0=None,
    dtheta0=None, dpsi0=None, tilt0=None,
    angle_nout=None,
    det_ref=None,
    units=None,
    plot_dets=None, nsort=None,
    tangent_to_rowland=None,
    miscut=None,
    pts=None,
    test_lamb_interv=None,
    contour=None,
    fs=None,
    cmap=None,
    vmin=None,
    vmax=None,
    ax=None,
):

    if cmap is None:
        # cmap = 'RdYlBu'
        cmap = plt.cm.viridis
    if nsort is None:
        nsort = 5
    if contour is None:
        errmin = np.nanmin(error_lambda)
        contour = [errmin + (np.nanmax(error_lambda) - errmin)/50.]
    if fs is None:
        fs = (6, 8)

    if ax is None:
        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(1, 1)
        ax = fig.add_subplot(gs[0, 0])
        ax.set_title('Mean focalization error\non detector')
        ax.set_xlabel('ddist (m)')
        ax.set_ylabel('di (m)')

    # plot error map function(ddist, di)
    extent = (ddist.min(), ddist.max(), di.min(), di.max())
    errmap = ax.imshow(
        error_lambda,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        origin='lower',
        extent=extent,
        interpolation='nearest',
    )
    cbar = plt.colorbar(
        errmap,
        label=f"error on lambda ({units})",
        orientation="vertical",
    )
    ax.contour(
        ddist,
        di,
        error_lambda,
        contour,
        colors='w',
        linestyles='-',
        linewidths=1.,
    )
    ax.contour(
        ddist,
        di,
        test_lamb_interv,
        contour,
        colors='yellow',
        linestyles='-',
        linewidths=1.,
    )

    # Computing detector with exact position of det_ref
    if det_ref:
        dpsi0bis = float(dpsi0)
        if tangent_to_rowland:
            dpsi0bis = dpsi0 - angle_nout

        detector_comp = cryst.get_detector_ideal(
            ddist=ddist0,
            di=di0,
            dj=dj0,
            dtheta=dtheta0,
            dpsi=dpsi0bis,
            tilt=tilt0,
            lamb=lamb,
            miscut=miscut,
            tangent_to_rowland=False,
        )
        detector_comp['outline'] = det_ref['outline']
        ax.plot(
            ddist0,
            di0,
            marker='x',
            ls='None',
            color='w',
        )

    if plot_dets:
        indsort = np.argsort(np.ravel(error_lambda))
        inddist = indsort % ddist.size
        inddi = indsort // ddist.size

        # plot nbr of dets on map "mean focalization error = f(ddist, di)"
        ax.plot(
            ddist[inddist[:nsort]],
            di[inddi[:nsort]],
            marker='x',
            ls='None',
            color='r',
        )

        # plot dets geometry with CrystalBragg_plot()
        if det_ref is not None:
            dax = cryst.plot(
                det=det_ref,
                pts=pts,
                color='black',
            )
            dax = cryst.plot(
                det=detector_comp,
                pts=pts,
                color='blue',
                element='ocv',
                dax=dax,
            )
            msg = (
                "Parameters of reference detector:\n"
                + "Center position in (x, y, z): ({})\n".format(
                    np.round(detector_comp['cent'], decimals=4)
                    )
                + "Translations (ddist, di, dj): ({}, {}, {}) [m]\n".format(
                    ddist0, di0, dj0,
                    )
                + "Rotations (dtheta, dpsi, tilt): ({}, {}, {}) [rad]\n".format(
                    dtheta0, dpsi0, tilt0,
                    )
            )
            print(msg)
            det = {}
            for ii in range(nsort):
                det[ii] = cryst.get_detector_ideal(
                    ddist=ddist[inddist[ii]],
                    di=di[inddi[ii]],
                    tangent_to_rowland=tangent_to_rowland,
                )
                det[ii]['outline'] = det_ref['outline']
                dax = cryst.plot(
                    det=det[ii],
                    pts=pts,
                    color='red',
                    element='oc',
                    dax=dax,
                )
                print(
                    "det: {}\n".format(det[ii])
                    + "\t ddist: {}\n".format(ddist[inddist[ii]])
                    + "\t di: {}\n".format(di[inddi[ii]])
                )
    return ax


# #################################################################
# #################################################################
#                   Ray tracing plot
# #################################################################
# #################################################################


# To be clarified
def CrystalBragg_plot_raytracing_from_lambpts(xi=None, xj=None, lamb=None,
                                              xi_bounds=None, xj_bounds=None,
                                              pts=None, ptscryst=None,
                                              ptsdet=None,
                                              det_cent=None, det_nout=None,
                                              det_ei=None, det_ej=None,
                                              cryst=None, proj=None,
                                              fs=None, ax=None, dmargin=None,
                                              wintit=None, tit=None,
                                              legend=None, draw=None):
    # Check
    assert xi.shape == xj.shape and xi.ndim == 3
    assert (isinstance(proj, list)
            and all([pp in ['det', '2d', '3d'] for pp in proj]))
    if legend is None or legend is True:
        legend = dict(bbox_to_anchor=(1.02, 1.), loc='upper left',
                      ncol=1, mode="expand", borderaxespad=0.,
                      prop={'size': 6})
    if wintit is None:
        wintit = _WINTIT
    if draw is None:
        draw = True

    # Prepare
    nlamb, npts, ndtheta = xi.shape
    det = np.array([[xi_bounds[0], xi_bounds[1], xi_bounds[1],
                     xi_bounds[0], xi_bounds[0]],
                    [xj_bounds[0], xj_bounds[0], xj_bounds[1],
                     xj_bounds[1], xj_bounds[0]]])
    lcol = ['r', 'g', 'b', 'm', 'y', 'c']
    lm = ['+', 'o', 'x', 's']
    lls = ['-', '--', ':', '-.']
    ncol, nm, nls = len(lcol), len(lm), len(lls)

    if '2d' in proj or '3d' in proj:
        pts = np.repeat(np.repeat(pts[:, None, :], nlamb, axis=1)[..., None],
                        ndtheta, axis=-1)[..., None]
        ptsall = np.concatenate(
            (
                pts,
                ptscryst[..., None],
                ptsdet[..., None],
                np.full((3, nlamb, npts, ndtheta, 1), np.nan),
            ),
            axis=-1,
        ).reshape((3, nlamb, npts, ndtheta*4))
        del pts, ptscryst, ptsdet
        if '2d' in proj:
            R = np.hypot(ptsall[0, ...], ptsall[1, ...])

    # --------
    # Plot
    lax = []
    if 'det' in proj:

        # Prepare
        if ax is None:
            if fs is None:
                fsi = (8, 6)
            else:
                fsi = fs
            if dmargin is None:
                dmargini = {'left': 0.1, 'right': 0.8,
                            'bottom': 0.1, 'top': 0.9,
                            'wspace': None, 'hspace': 0.4}
            else:
                dmargini = dmargin
            if tit is None:
                titi = False
            else:
                titi = tit
            fig = plt.figure(figsize=fsi)
            gs = gridspec.GridSpec(1, 1, **dmargini)
            axi = fig.add_subplot(
                gs[0, 0], aspect='equal', adjustable='datalim',
            )
            axi.set_xlabel(r'$x_i$ (m)')
            axi.set_ylabel(r'$x_j$ (m)')
        else:
            axi = ax

        # plot
        axi.plot(det[0, :], det[1, :], ls='-', lw=1., c='k')
        for pp in range(npts):
            for ll in range(nlamb):
                lab = (
                    r'pts {} - '.format(pp)
                    + r'$\lambda$' + ' = {:6.3f} A'.format(lamb[ll]*1.e10)
                )
                axi.plot(
                    xi[ll, pp, :], xj[ll, pp, :],
                    ls='None', marker=lm[ll % nm],
                    c=lcol[pp % ncol], label=lab,
                )

        # decorate
        if legend is not False:
            axi.legend(**legend)
        if wintit is not False:
            axi.figure.canvas.manager.set_window_title(wintit)
        if titi is not False:
            axi.figure.suptitle(titi, size=14, weight='bold')
        if draw:
            axi.figure.canvas.draw()
        lax.append(axi)

    if '2d' in proj:

        # Prepare
        if tit is None:
            titi = False
        else:
            titi = tit

        # plot
        dax = cryst.plot(lax=ax, proj='all',
                         det_cent=det_cent, det_nout=det_nout,
                         det_ei=det_ei, det_ej=det_ej, draw=False)
        for pp in range(npts):
            for ll in range(nlamb):
                lab = (r'pts {} - '.format(pp)
                       + r'$\lambda$'+' = {:6.3f} A'.format(lamb[ll]*1.e10))
                dax['cross'].plot(
                    R[ll, pp, :], ptsall[2, ll, pp, :],
                    ls=lls[ll % nls], color=lcol[pp % ncol],
                    label=lab,
                )
                dax['hor'].plot(
                    ptsall[0, ll, pp, :], ptsall[1, ll, pp, :],
                    ls=lls[ll % nls], color=lcol[pp % ncol], label=lab,
                )
        # decorate
        if legend is not False:
            dax['cross'].legend(**legend)
        if wintit is not False:
            dax['cross'].figure.canvas.manager.set_window_title(wintit)
        if titi is not False:
            dax['cross'].figure.suptitle(titi, size=14, weight='bold')
        if draw:
            dax['cross'].figure.canvas.draw()
        lax.append(dax['cross'])
        lax.append(dax['hor'])

    return lax


# #################################################################
# #################################################################
#                   plot plasma domain from lamb
# #################################################################


def CrystalBragg_plot_plasma_domain_at_lamb(
    cryst=None,
    det=None,
    xixj_lim=None,
    config=None,
    lamb=None,
    pts=None,
    reseff=None,
    lambok=None,
    dax=None,
    plot_as=None,
    lcolor=None,
):

    # -------------
    # check inputs

    if plot_as is None:
        plot_as = 'poly'
    if plot_as not in ['pts', 'poly']:
        msg = (
            "Arg plot_as not in allowed values!\n"
            f"\t- allowed: {['pts', 'poly']}\n"
            f"\t- plot_as: {plot_as}"
        )
        raise Exception(msg)

    nlamb = lamb.size
    if lcolor is None:
        lcolor = [ss['color'] for ss in mpl.rcParams['axes.prop_cycle']]
    if not isinstance(lcolor, list):
        lcolor = [lcolor]
    if not isinstance(lcolor, list):
        msg = (
            f"Arg lcolor must be a list of mpl colors\n"
            f"\t- Provided: {lcolor}"
        )
        raise Exception(msg)

    # -------------
    # prepare data

    R_u = np.unique(pts[0, :])
    Z_u = np.unique(pts[1, :])
    nR, nZ = R_u.size, Z_u.size
    extent = (R_u.min(), R_u.max(), Z_u.min(), Z_u.max())

    indR = [pts[0, :] == rr for rr in R_u]
    indZ = [pts[1, :] == zz for zz in Z_u]
    phi_per_R = [np.unique(pts[2, pts[0, :] == rr]) for rr in R_u]
    nphi_per_R = np.array([ppr.size for ppr in phi_per_R])
    nphimax = np.max(nphi_per_R)
    indPhi = [
        [pts[2, :] == phi for phi in phi_per_R[ii]]
        for ii in range(R_u.size)
    ]

    cross = np.zeros((nlamb, R_u.size, Z_u.size), dtype=bool)
    hor = np.zeros((nlamb, R_u.size, nphimax), dtype=bool)
    horR = np.full((R_u.size, nphimax), np.nan)
    horPhi = np.full((R_u.size, nphimax), np.nan)

    for ii, rr in enumerate(R_u):

        for jj, zz in enumerate(Z_u):
            ind = indR[ii] & indZ[jj]
            for kk in range(nlamb):
                cross[kk, ii, jj] = np.any(lambok[kk, ind])

        horR[ii, :] = rr
        horPhi[ii, :nphi_per_R[ii]] = phi_per_R[ii]
        for jj, pp in enumerate(phi_per_R[ii]):
            ind = indR[ii] & indPhi[ii][jj]
            for kk in range(nlamb):
                hor[kk, ii, jj] = np.any(lambok[kk, ind])

    # ------------
    # get contours

    if plot_as == 'poly':
        # Add envelop
        dR = (R_u[-1] - R_u[0])/R_u.size
        dZ = (Z_u[-1] - Z_u[0])/Z_u.size
        x = np.tile(np.r_[R_u[0] - dR, R_u, R_u[-1] + dR], (nZ + 2, 1))
        y = np.tile(np.r_[Z_u[0] - dZ, Z_u, Z_u[-1] + dZ], (nR + 2, 1)).T
        z = np.zeros(x.shape, dtype=float)

        # see https://github.com/matplotlib/matplotlib/blob/main/src/_contour.h
        cont_cross = [None for ll in lamb]
        for kk, ll in enumerate(lamb):
            z[1:-1, 1:-1] = cross[kk, ...].T
            cont_raw = contour_generator(
                x=x,
                y=y,
                z=z,
                name='serial',
                corner_mask=None,
                line_type='Separate',
                fill_type=None,
                chunk_size=None,
                chunk_count=None,
                total_chunk_count=None,
                quad_as_tri=True,       # for sub-mesh precision
                # z_interp=<ZInterp.Linear: 1>,
                thread_count=0,
            ).lines(0.5)
            assert all([pp.ndim == 2  and pp.shape[1] == 2 for pp in cont_raw])
            cont_cross[kk] = PatchCollection(
                [plt.Polygon(pp) for pp in cont_raw],
                color=lcolor[kk % nlamb],
                alpha=0.4,
            )
    else:
        R_cross = np.tile(R_u, (Z_u.size, 1)).T
        Z_cross = np.tile(Z_u, (R_u.size, 1))

    # -------------
    # plot context

    if dax is None:
        lax = None
    else:
        lax = [dax['cross'], dax['hor']]
    lax = config.plot(lax=lax)
    dax = {'cross': lax[0], 'hor': lax[1]}

    if plot_as == 'pts':
        for kk, ll in enumerate(lamb):
            dax['cross'].plot(
                R_cross[cross[kk, ...]],
                Z_cross[cross[kk, ...]],
                alpha=0.4,
                ls='None',
                marker='o',
                ms=6,
                markeredgecolor='None',
                color=lcolor[kk % nlamb],
            )
            # dax['cross'].plot(cont_cross[:, 0], cont_cross[:, 1])
            dax['hor'].plot(
                horR[hor[kk, ...]]*np.cos(horPhi[hor[kk, ...]]),
                horR[hor[kk, ...]]*np.sin(horPhi[hor[kk, ...]]),
                color=lcolor[kk % nlamb],
                alpha=0.4,
                ls='None',
                marker='o',
                ms=6,
                markeredgecolor='None',
            )
    else:
        for kk, ll in enumerate(lamb):
            dax['cross'].add_collection(
                cont_cross[kk],
            )
            dax['hor'].plot(
                horR[hor[kk, ...]]*np.cos(horPhi[hor[kk, ...]]),
                horR[hor[kk, ...]]*np.sin(horPhi[hor[kk, ...]]),
                color=lcolor[kk % nlamb],
                alpha=0.4,
                ls='None',
                marker='o',
                ms=6,
                markeredgecolor='None',
            )

    dax = cryst.plot(
        det=det,
        dax=dax,
        proj=['cross', 'hor'],
    )
    return dax


# #################################################################
# #################################################################
#               plot synthetic signal from emissivity
# #################################################################


def CrystalBragg_plot_signal_from_emissivity(
    cryst=None,
    det=None,
    xixj_lim=None,
    config=None,
    lamb=None,
    pts=None,
    reseff=None,
    xi=None,
    xj=None,
    val=None,
    lambok=None,
    binning=None,
    binned=None,
    # plotting
    vmin=None,
    vmax=None,
    vmin_bin=None,
    vmax_bin=None,
    cmap=None,
    dax=None,
    fs=None,
    dmargin=None,
    tit=None,
):

    # -------------
    # check inputs

    nlamb = lamb.size

    if fs is None:
        fs = (12, 8)
    if dmargin is None:
        dmargin = {'left': 0.1, 'right': 0.95,
                   'bottom': 0.1, 'top': 0.9,
                   'wspace': 0.3, 'hspace': 0.2}
    if tit is None:
        tit = False

    if vmin is None:
        vmin = 0
    if vmax is None:
        vmax = np.nanmax(val)
    if vmin_bin is None:
        vmin_bin = 0
    if vmax_bin is None and binning is not False:
        vmax_bin = np.nanmax(binned)
    if cmap is None:
        cmap = plt.cm.viridis

    # -------------
    # prepare data

    R_u = np.unique(pts[0, :])
    Z_u = np.unique(pts[1, :])
    nR, nZ = R_u.size, Z_u.size
    extent = (R_u.min(), R_u.max(), Z_u.min(), Z_u.max())

    indR = [pts[0, :] == rr for rr in R_u]
    indZ = [pts[1, :] == zz for zz in Z_u]
    phi_per_R = [np.unique(pts[2, pts[0, :] == rr]) for rr in R_u]
    nphi_per_R = np.array([ppr.size for ppr in phi_per_R])
    nphimax = np.max(nphi_per_R)
    indPhi = [
        [pts[2, :] == phi for phi in phi_per_R[ii]]
        for ii in range(R_u.size)
    ]

    cross = np.zeros((R_u.size, Z_u.size), dtype=bool)
    hor = np.zeros((R_u.size, nphimax), dtype=bool)
    horR = np.full((R_u.size, nphimax), np.nan)
    horPhi = np.full((R_u.size, nphimax), np.nan)

    for ii, rr in enumerate(R_u):
        for jj, zz in enumerate(Z_u):
            ind = indR[ii] & indZ[jj]
            cross[ii, jj] = np.any(lambok[:, ind])

        horR[ii, :] = rr
        horPhi[ii, :nphi_per_R[ii]] = phi_per_R[ii]
        for jj, pp in enumerate(phi_per_R[ii]):
            ind = indR[ii] & indPhi[ii][jj]
            hor[ii, jj] = np.any(lambok[:, ind])

    # ------------
    # get contours

    # plot_as == 'poly':
    # Add envelop
    dR = (R_u[-1] - R_u[0])/R_u.size
    dZ = (Z_u[-1] - Z_u[0])/Z_u.size
    x = np.tile(np.r_[R_u[0] - dR, R_u, R_u[-1] + dR], (nZ + 2, 1))
    y = np.tile(np.r_[Z_u[0] - dZ, Z_u, Z_u[-1] + dZ], (nR + 2, 1)).T
    z = np.zeros(x.shape, dtype=float)

    # see https://github.com/matplotlib/matplotlib/blob/main/src/_contour.h
    z[1:-1, 1:-1] = cross.T
    cont_raw = contour_generator(
        x=x,
        y=y,
        z=z,
        name='serial',
        corner_mask=None,
        line_type='Separate',
        fill_type=None,
        chunk_size=None,
        chunk_count=None,
        total_chunk_count=None,
        quad_as_tri=True,       # for sub-mesh precision
        # z_interp=<ZInterp.Linear: 1>,
        thread_count=0,
    ).lines(0.5)
    assert all([pp.ndim == 2 and pp.shape[1] == 2 for pp in cont_raw])
    cont_cross = PatchCollection(
        [plt.Polygon(pp) for pp in cont_raw],
        color='k',
        alpha=0.4,
    )

    # -----------
    # binned data

    if binning is not False:
        extent = (binning[0][0], binning[0][-1], binning[1][0], binning[1][-1])

    # -------------
    # plot context

    if dax is None:

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(2, 3, **dmargin)

        ax0 = fig.add_subplot(
            gs[0, 0], aspect='equal', adjustable='datalim',
        )
        ax1 = fig.add_subplot(
            gs[1, 0], aspect='equal', adjustable='datalim',
        )
        ax2 = fig.add_subplot(
            gs[:, 1], aspect='equal',
        )
        ax3 = fig.add_subplot(
            gs[:, 2], sharex=ax2, sharey=ax2,
        )

        ax0.set_xlabel(r'$R$ (m)')
        ax0.set_ylabel(r'$Z$ (m)')

        ax1.set_xlabel(r'$x$ (m)')
        ax1.set_ylabel(r'$y$ (m)')

        ax2.set_xlabel(r'$x_i$ (m)')
        ax2.set_ylabel(r'$x_j$ (m)')

        ax3.set_xlabel(r'$x_i$ (m)')
        ax3.set_ylabel(r'$x_j$ (m)')

        dax = {
            'cross': {'ax': ax0, 'type': 'cross'},
            'hor': {'ax': ax1, 'type': 'hor'},
            'img1': {'ax': ax2, 'type': 'img'},
            'img2': {'ax': ax3, 'type': 'img'},
        }

    lax = config.plot(lax=[dax.get('cross')['ax'], dax.get('hor')['ax']])

    # ----------------
    # actual plotting

    k0 = 'cross'
    if dax.get(k0) is not None:
        ax = dax[k0]['ax']
        ax.add_collection(cont_cross)

    k0 = 'hor'
    if dax.get(k0) is not None:
        ax = dax[k0]['ax']
        ax.plot(
            horR[hor]*np.cos(horPhi[hor]),
            horR[hor]*np.sin(horPhi[hor]),
            color='k',
            alpha=0.4,
            ls='None',
            marker='o',
            ms=6,
            markeredgecolor='None',
        )

    k0 = 'img1'
    if dax.get(k0) is not None:
        ax = dax[k0]['ax']
        iok = np.isfinite(val)
        ax.scatter(
            xi[iok].ravel(),
            xj[iok].ravel(),
            s=2,
            c=val[iok].ravel(),
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            marker='.',
        )
        ax.plot(
            det['outline'][0, :],
            det['outline'][1, :],
            c='k',
            ls='-',
            lw=1.,
        )

    k0 = 'img2'
    if dax.get(k0) is not None and binning is not False:
        ax = dax[k0]['ax']
        ax.imshow(
            binned.T,
            origin='lower',
            extent=extent,
            interpolation='nearest',
            cmap=cmap,
            vmin=vmin_bin,
            vmax=vmax_bin,
        )
        ax.plot(
            det['outline'][0, :],
            det['outline'][1, :],
            c='k',
            ls='-',
            lw=1.,
        )

    return dax



# coding utf-8

# Built-in
import itertools as itt
import warnings

# Common
import numpy as np
# import scipy.integrate as scpinteg
import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib as mpl
import matplotlib.transforms as transforms
import matplotlib.lines as mlines
import matplotlib.colors as mcolors
# from mpl_toolkits.axes_grid1 import make_axes_locatable
import datastock as ds


# tofu
from tofu.version import __version__
from .. import utils as utils
from . import _def as _def
from . import _generic_check
from . import _DataCollection_plot_text


__all__ = [
    'plot_TimeTraceColl',
    'plot_axvline',
]


# __author_email__ = 'didier.vezinet@cea.fr'
__github = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = 'tofu-%s        report issues / requests at {}'.format(
    __version__, __github,
)
_nchMax, _ntMax, _nfMax, _nlbdMax = 4, 3, 3, 3
_fontsize = 8
_labelpad = 0
_lls = ['-', '--', '-.', ':']
_lct = [plt.cm.tab20.colors[ii] for ii in [0, 2, 4, 1, 3, 5]]
_lcch = [plt.cm.tab20.colors[ii] for ii in [6, 8, 10, 7, 9, 11]]
_lclbd = [plt.cm.tab20.colors[ii] for ii in [12, 16, 18, 13, 17, 19]]
_lcm = _lclbd
_cbck = (0.8, 0.8, 0.8)
_dmarker = {'ax': 'o', 'x': 'x'}


_OVERHEAD = True
_CROSS = True
_DRAW = True
_CONNECT = True
_AXGRID = False
_LIB = 'mpl'
_BCKCOLOR = 'w'

_LCOLOR_DICT = [
    [
        'tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple',
        'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan',
    ],
    ['r', 'g', 'b'],
    ['m', 'y', 'c'],
]


# #############################################################################
# #############################################################################
#                       generic entry point
# #############################################################################


def plot_as_array(
    # parameters
    coll=None,
    key=None,
    ind=None,
    vmin=None,
    vmax=None,
    cmap=None,
    ymin=None,
    ymax=None,
    aspect=None,
    nmax=None,
    color_dict=None,
    dinc=None,
    lkeys=None,
    bstr_dict=None,
    # figure-specific
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    connect=None,
    inplace=None,
):

    # ------------
    #  ceck inputs

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=coll.ddata.keys(),
    )
    ndim = coll._ddata[key]['data'].ndim

    inplace = ds._generic_check._check_var(
        inplace, 'inplace',
        types=bool,
        default=False,
    )

    if inplace:
        coll2 = coll
    else:
        lk0 = list(itt.chain.from_iterable([
            [
                k0 for k0, v0 in coll._ddata.items()
                if v0['ref'] == (rr,)
            ]
            for rr in coll._ddata[key]['ref']
        ]))
        coll2 = coll.extract([key] + lk0)

    # -------------------------
    #  call appropriate routine

    if ndim == 1:
        return plot_as_array_1d(
            # parameters
            coll=coll2,
            key=key,
            ind=ind,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            ymin=ymin,
            ymax=ymax,
            aspect=aspect,
            nmax=nmax,
            color_dict=color_dict,
            dinc=dinc,
            lkeys=lkeys,
            bstr_dict=bstr_dict,
            # figure-specific
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            connect=connect,
        )

    elif ndim == 2:
        return plot_as_array_2d(
            # parameters
            coll=coll2,
            key=key,
            ind=ind,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            ymin=ymin,
            ymax=ymax,
            aspect=aspect,
            nmax=nmax,
            color_dict=color_dict,
            dinc=dinc,
            lkeys=lkeys,
            bstr_dict=bstr_dict,
            # figure-specific
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            connect=connect,
        )

    elif ndim == 3:
        return plot_as_array_3d(
            # parameters
            coll=coll2,
            key=key,
            ind=ind,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            ymin=ymin,
            ymax=ymax,
            aspect=aspect,
            nmax=nmax,
            color_dict=color_dict,
            dinc=dinc,
            # figure-specific
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            connect=connect,
        )


# #############################################################################
# #############################################################################
#                       check
# #############################################################################


def _plot_as_array_check(
    ndim=None,
    coll=None,
    key=None,
    ind=None,
    cmap=None,
    vmin=None,
    vmax=None,
    ymin=None,
    ymax=None,
    aspect=None,
    nmax=None,
    color_dict=None,
    dcolorbar=None,
    dleg=None,
    data=None,
    connect=None,
    groups=None,
):

    # ind
    ind = ds._generic_check._check_var(
        ind, 'ind',
        default=[0 for ii in range(ndim)],
        types=(list, tuple, np.ndarray),
    )
    c0 = (
        len(ind) == ndim
        and all([
            np.isscalar(ii) and isinstance(ii, (int, np.integer))
            for ii in ind
        ])
    )
    if not c0:
        msg = (
            "Arg ind must be an iterable of 2 integer indices!\n"
            f"Provided: {ind}"
        )
        raise Exception(msg)

    # cmap
    if cmap is None or vmin is None or vmax is None:
        if isinstance(coll.ddata[key]['data'], np.ndarray):
            nanmax = np.nanmax(coll.ddata[key]['data'])
            nanmin = np.nanmin(coll.ddata[key]['data'])
        else:
            nanmax = coll.ddata[key]['data'].max()
            nanmin = coll.ddata[key]['data'].min()
        diverging = nanmin * nanmax <= 0

    if cmap is None:
        if diverging:
            cmap = 'seismic'
        else:
            cmap = 'viridis'

    # vmin, vmax
    if vmin is None and diverging:
        vmin = -max(abs(nanmin), nanmax)
    if vmax is None and diverging:
        vmax = max(abs(nanmin), nanmax)

    # vmin, vmax
    if ymin is None:
        ymin = vmin
    if ymax is None:
        ymax = vmax

    # aspect
    aspect = ds._generic_check._check_var(
        aspect, 'aspect',
        default='equal',
        types=str,
        allowed=['auto', 'equal'],
    )

    # nmax
    nmax = ds._generic_check._check_var(
        nmax, 'nmax',
        default=3,
        types=int,
    )

    # color_dict
    cdef = {
        k0: _LCOLOR_DICT[0] for ii, k0 in enumerate(groups)
    }
    color_dict = ds._generic_check._check_var(
        color_dict, 'color_dict',
        default=cdef,
        types=dict,
    )
    dout = {
        k0: str(v0)
        for k0, v0 in color_dict.items()
        if not (
            isinstance(k0, str)
            and k0 in groups
            and isinstance(v0, list)
            and all([mcolors.is_color_like(v1) for v1 in v0])
        )
    }
    if len(dout) > 0:
        lstr = [f"{k0}: {v0}" for k0, v0 in dout.items()]
        msg = (
            "The following entries of color_dict are invalid"
        )

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = ds._generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    # connect
    connect = ds._generic_check._check_var(
        connect, 'connect',
        default=True,
        types=bool,
    )

    return (
        key, ind,
        cmap, vmin, vmax,
        ymin, ymax,
        aspect, nmax,
        color_dict,
        dcolorbar, dleg, connect,
    )


# #############################################################################
# #############################################################################
#                       plot_as_array: 1d
# #############################################################################


def plot_as_array_1d(
    # parameters
    coll=None,
    key=None,
    ind=None,
    vmin=None,
    vmax=None,
    cmap=None,
    ymin=None,
    ymax=None,
    aspect=None,
    nmax=None,
    color_dict=None,
    dinc=None,
    lkeys=None,
    bstr_dict=None,
    # figure-specific
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    connect=None,
):

    # --------------
    # check input

    groups = ['ref']
    (
        key, ind,
        cmap, vmin, vmax,
        ymin, ymax,
        aspect, nmax,
        color_dict,
        dcolorbar, dleg, connect,
    ) = _plot_as_array_check(
        ndim=1,
        coll=coll,
        key=key,
        ind=ind,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        ymin=ymin,
        ymax=ymax,
        aspect=aspect,
        nmax=nmax,
        color_dict=color_dict,
        dcolorbar=dcolorbar,
        dleg=dleg,
        connect=connect,
        groups=groups,
    )

    # --------------
    #  Prepare data

    data = coll.ddata[key]['data']
    if hasattr(data, 'nnz'):
        data = data.toarray()
    assert data.ndim == len(coll.ddata[key]['ref']) == 1
    n0, = data.shape

    ref = coll._ddata[key]['ref'][0]
    units = coll._ddata[key]['units']
    lab0 = f'ind ({ref})'
    lab1 = f'{key} ({units})'

    # --------------
    # plot - prepare

    if dax is None:

        if fs is None:
            fs = (12, 8)

        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.95,
                'bottom': 0.05, 'top': 0.90,
                'hspace': 0.15, 'wspace': 0.2,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=4, nrows=1, **dmargin)

        ax0 = fig.add_subplot(gs[0, :3], aspect='auto')
        ax0.set_xlabel(lab0)
        ax0.set_ylabel(lab1)
        ax0.set_title(key, size=14, fontweight='bold')

        ax1 = fig.add_subplot(gs[0, 3], frameon=False)
        ax1.set_xticks([])
        ax1.set_yticks([])

        dax = {
            'misc': {'handle': ax0, 'type': 'misc'},
            'text': {'handle': ax1, 'type': 'text'},
        }

    dax = _generic_check._check_dax(dax=dax, main='misc')

    # ---------------
    # plot fixed part

    axtype = 'misc'
    lkax = [kk for kk, vv in dax.items() if vv['type'] == axtype]
    for kax in lkax:
        ax = dax[kax]['handle']

        ax.plot(
            data,
            color='k',
            marker='.',
            ms=6,
        )

        # plt.colorbar(im, ax=ax, **dcolorbar)
        if dleg is not False:
            ax.legend(**dleg)

    # ----------------
    # define and set dgroup

    dgroup = {
        ref: {
            'ref': [ref],
            'data': ['index'],
            'nmax': nmax,
        },
    }

    # ----------------
    # plot mobile part

    axtype = 'misc'
    lkax = [kk for kk, vv in dax.items() if vv['type'] == axtype]
    for kax in lkax:
        ax = dax[kax]['handle']

        # ind0, ind1
        for ii in range(nmax):
            lv = ax.axvline(ind[0], c=color_dict['ref'][ii], lw=1., ls='-')

            # update coll
            kv = f'v{ii:02.0f}'
            coll.add_mobile(
                key=kv,
                handle=lv,
                ref=ref,
                data='index',
                dtype='xdata',
                ax=kax,
                ind=ii,
            )

        dax[kax].update(refx=[ref])

    # ---------
    # add text

    kax = 'text'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _DataCollection_plot_text.plot_text(
            coll=coll,
            kax=kax,
            ax=ax,
            ref=ref,
            group='ref',
            ind=ind[0],
            lkeys=lkeys,
            nmax=nmax,
            color_dict=color_dict,
            bstr_dict=bstr_dict,
        )

    # add axes
    for kax in dax.keys():
        coll.add_axes(key=kax, **dax[kax])

    # increment dict

    coll.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)

    # connect
    if connect is True:
        coll.connect()

    return coll


# #############################################################################
# #############################################################################
#                       plot_as_array: 2d
# #############################################################################


def plot_as_array_2d(
    # parameters
    coll=None,
    key=None,
    ind=None,
    vmin=None,
    vmax=None,
    cmap=None,
    ymin=None,
    ymax=None,
    aspect=None,
    nmax=None,
    color_dict=None,
    dinc=None,
    lkeys=None,
    bstr_dict=None,
    # figure-specific
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    connect=None,
):

    # --------------
    # check input

    groups = ['hor', 'vert']
    (
        key, ind,
        cmap, vmin, vmax,
        ymin, ymax,
        aspect, nmax,
        color_dict,
        dcolorbar, dleg, connect,
    ) = _plot_as_array_check(
        ndim=2,
        coll=coll,
        key=key,
        ind=ind,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        ymin=ymin,
        ymax=ymax,
        aspect=aspect,
        nmax=nmax,
        color_dict=color_dict,
        dcolorbar=dcolorbar,
        dleg=dleg,
        connect=connect,
        groups=groups,
    )

    # --------------
    #  Prepare data

    data = coll.ddata[key]['data']
    if hasattr(data, 'nnz'):
        data = data.toarray()
    assert data.ndim == len(coll.ddata[key]['ref']) == 2
    n0, n1 = data.shape
    extent = (-0.5, n1 - 0.5, -0.5, n0 - 0.5)

    ref0, ref1 = coll.ddata[key]['ref']
    lab0 = f'ind0 ({ref0})'
    lab1 = f'ind1 ({ref1})'

    # --------------
    # plot - prepare

    if dax is None:

        if fs is None:
            fs = (14, 8)

        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.95,
                'bottom': 0.06, 'top': 0.90,
                'hspace': 0.2, 'wspace': 0.3,
            }

        fig = plt.figure(figsize=fs)
        fig.suptitle(key, size=14, fontweight='bold')
        gs = gridspec.GridSpec(ncols=4, nrows=6, **dmargin)

        ax0 = fig.add_subplot(gs[:4, :2], aspect='auto')
        ax0.set_ylabel(lab0)
        ax0.set_xlabel(lab1)
        ax0.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax0.xaxis.set_label_position('top')

        ax1 = fig.add_subplot(gs[:4, 2], sharey=ax0)
        ax1.set_xlabel('data')
        ax1.set_ylabel(lab0)
        ax1.tick_params(
            axis="y",
            left=False, right=True,
            labelleft=False, labelright=True,
        )
        ax1.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax1.yaxis.set_label_position('right')
        ax1.xaxis.set_label_position('top')

        ax2 = fig.add_subplot(gs[4:, :2], sharex=ax0)
        ax2.set_ylabel('data')
        ax2.set_xlabel(lab1)

        ax1.set_xlim(ymin, ymax)
        ax2.set_ylim(ymin, ymax)

        # axes for text
        ax3 = fig.add_subplot(gs[:3, 3], frameon=False)
        ax3.set_xticks([])
        ax3.set_yticks([])
        ax4 = fig.add_subplot(gs[3:, 3], frameon=False)
        ax4.set_xticks([])
        ax4.set_yticks([])

        axy = ax1.get_position().bounds
        ax5 = fig.add_axes(
            [axy[0], axy[1]-0.02, 0.9*axy[2], 0.02],
            frameon=False,
        )
        ax5.set_xticks([])
        ax5.set_yticks([])
        axy = ax2.get_position().bounds
        dy = 0.1*(axy[3] - axy[1])
        ax6 = fig.add_axes(
            [axy[0] + axy[2], axy[1] + dy, 0.03, axy[3] - dy],
            frameon=False,
        )
        ax6.set_xticks([])
        ax6.set_yticks([])

        dax = {
            # data
            'matrix': {'handle': ax0, 'type': 'matrix', 'inverty': True},
            'vertical': {'handle': ax1, 'type': 'misc', 'inverty': True},
            'horizontal': {'handle': ax2, 'type': 'misc'},
            # text
            'text0': {'handle': ax3, 'type': 'text'},
            'text1': {'handle': ax4, 'type': 'text'},
            'vertical_text': {'handle': ax5, 'type': 'text'},
            'horizontal_text': {'handle': ax6, 'type': 'text'},
        }

    dax = _generic_check._check_dax(dax=dax, main='matrix')

    # ---------------
    # plot fixed part

    axtype = 'matrix'
    lkax = [kk for kk, vv in dax.items() if vv['type'] == axtype]
    for kax in lkax:
        ax = dax[kax]['handle']

        im = ax.imshow(
            data,
            extent=extent,
            interpolation='nearest',
            origin='lower',
            aspect=aspect,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )
        ax.invert_yaxis()

    # ----------------
    # define and set dgroup

    dgroup = {
        'hor': {
            'ref': [ref0],
            'data': ['index'],
            'nmax': nmax,
        },
        'vert': {
            'ref': [ref1],
            'data': ['index'],
            'nmax': nmax,
        },
    }

    # ----------------
    # plot mobile part

    kax = 'matrix'
    for kax in lkax:
        ax = dax[kax]['handle']

        # ind0, ind1
        for ii in range(nmax):
            lh = ax.axhline(ind[0], c=color_dict['hor'][ii], lw=1., ls='-')
            lv = ax.axvline(ind[1], c=color_dict['vert'][ii], lw=1., ls='-')

            # update coll
            kh = f'h{ii:02.0f}'
            kv = f'v{ii:02.0f}'
            coll.add_mobile(
                key=kh,
                handle=lh,
                ref=ref0,
                data='index',
                dtype='ydata',
                ax=kax,
                ind=ii,
            )
            coll.add_mobile(
                key=kv,
                handle=lv,
                ref=ref1,
                data='index',
                dtype='xdata',
                ax=kax,
                ind=ii,
            )

        dax[kax].update(refx=[ref1], refy=[ref0])

    kax = 'vertical'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        for ii in range(nmax):
            l0, = ax.plot(
                data[:, ind[1]],
                np.arange(0, n0),
                ls='-',
                marker='.',
                lw=1.,
                color=color_dict['vert'][ii],
                label=f'ind0 = {ind[0]}',
            )

            km = f'vprof{ii:02.0f}'
            coll.add_mobile(
                key=km,
                handle=l0,
                ref=(ref1,),
                data=key,
                dtype='xdata',
                ax=kax,
                ind=ii,
            )

            l0 = ax.axhline(
                ind[1],
                c=color_dict['hor'][ii],
            )
            km = f'lh-v{ii:02.0f}'
            coll.add_mobile(
                key=km,
                handle=l0,
                ref=(ref0,),
                data='index',
                dtype='ydata',
                ax=kax,
                ind=ii,
            )

        dax[kax].update(refy=[ref0])

    kax = 'horizontal'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        for ii in range(nmax):
            l1, = ax.plot(
                np.arange(0, n1),
                data[ind[0], :],
                ls='-',
                marker='.',
                lw=1.,
                color=color_dict['hor'][ii],
                label=f'ind1 = {ind[1]}',
            )

            km = f'hprof{ii:02.0f}'
            coll.add_mobile(
                key=km,
                handle=l1,
                ref=(ref0,),
                data=key,
                dtype='ydata',
                ax=kax,
                ind=ii,
            )

            l0 = ax.axvline(
                ind[0],
                c=color_dict['vert'][ii],
            )
            km = f'lv-h{ii:02.0f}'
            coll.add_mobile(
                key=km,
                handle=l0,
                ref=(ref1,),
                data='index',
                dtype='xdata',
                ax=kax,
                ind=ii,
            )

        dax[kax].update(refx=[ref1])

    # ---------
    # add text

    kax = 'vertical_text'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        for ii in range(nmax):
            ht = ax.text(
                (ii + 1)/nmax,
                0,
                '',
                horizontalalignment='left',
                verticalalignment='bottom',
                transform=ax.transAxes,
                size=10,
                color=color_dict['hor'][ii],
                fontweight='bold',
            )
            kt = f'vt0-v{ii:02.0f}'
            coll.add_mobile(
                key=kt,
                handle=ht,
                ref=(ref1,),
                data='index',
                dtype='txt',
                bstr='{0}',
                ax=kax,
                ind=ii,
            )

        ax.text(
            0, 0,
            'ind1 = ',
            horizontalalignment='left',
            verticalalignment='bottom',
            transform=ax.transAxes,
            size=10,
            fontweight='bold',
        )

    kax = 'horizontal_text'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        for ii in range(nmax):
            ht = ax.text(
                0,
                1 - (ii + 1)/nmax,
                '',
                horizontalalignment='left',
                verticalalignment='top',
                transform=ax.transAxes,
                size=10,
                color=color_dict['hor'][ii],
                fontweight='bold',
            )
            kt = f'ht0-v{ii:02.0f}'
            coll.add_mobile(
                key=kt,
                handle=ht,
                ref=(ref0,),
                data='index',
                dtype='txt',
                bstr='{0}',
                ax=kax,
                ind=ii,
            )

        ax.text(
            0, 1,
            'ind0 = ',
            horizontalalignment='left',
            verticalalignment='top',
            transform=ax.transAxes,
            size=10,
            fontweight='bold',
        )

    kax = 'text0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _DataCollection_plot_text.plot_text(
            coll=coll,
            kax=kax,
            ax=ax,
            ref=ref0,
            group='hor',
            ind=ind[0],
            lkeys=lkeys,
            nmax=nmax,
            color_dict=color_dict,
            bstr_dict=bstr_dict,
        )

    kax = 'text1'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _DataCollection_plot_text.plot_text(
            coll=coll,
            kax=kax,
            ax=ax,
            ref=ref1,
            group='vert',
            ind=ind[1],
            lkeys=lkeys,
            nmax=nmax,
            color_dict=color_dict,
            bstr_dict=bstr_dict,
        )

    # --------------------------
    # add axes and interactivity

    # add axes
    for kax in dax.keys():
        coll.add_axes(key=kax, **dax[kax])

    # increment dict

    coll.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)

    # connect
    if connect is True:
        coll.connect()

    return coll


# #############################################################################
# #############################################################################
#                       plot_as_array: 3d
# #############################################################################


def plot_as_array_3d(
    # parameters
    coll=None,
    key=None,
    ind=None,
    vmin=None,
    vmax=None,
    cmap=None,
    ymin=None,
    ymax=None,
    aspect=None,
    nmax=None,
    color_dict=None,
    dinc=None,
    # figure-specific
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    connect=None,
):

    msg = "Will be available in the next version"
    raise NotImplementedError(msg)

    # --------------
    # check input

    groups = ['i0', 'i1', 'i2']
    (
        key, ind,
        cmap, vmin, vmax,
        ymin, ymax,
        aspect, nmax,
        color_dict,
        dcolorbar, dleg, connect,
    ) = _plot_as_array_check(
        ndim=3,
        coll=coll,
        key=key,
        ind=ind,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        ymin=ymin,
        ymax=ymax,
        aspect=aspect,
        nmax=nmax,
        color_dict=color_dict,
        dcolorbar=dcolorbar,
        dleg=dleg,
        connect=connect,
        groups=groups,
    )
    nmax = 1

    # --------------
    #  Prepare data

    data = coll.ddata[key]['data']
    if hasattr(data, 'nnz'):
        data = data.toarray()
    assert data.ndim == len(coll.ddata[key]['ref']) == 3
    n0, n1, n2 = data.shape
    extent0 = (-0.5, n1 - 0.5, -0.5, n0 - 0.5)
    extent1 = (-0.5, n2 - 0.5, -0.5, n0 - 0.5)
    extent2 = (-0.5, n2 - 0.5, -0.5, n1 - 0.5)

    ref0, ref1, ref2 = coll.ddata[key]['ref']
    lab00 = f'ind0 ({ref0})'
    lab01 = f'ind1 ({ref1})'
    lab10 = f'ind0 ({ref0})'
    lab11 = f'ind2 ({ref2})'
    lab20 = f'ind1 ({ref1})'
    lab21 = f'ind2 ({ref2})'

    # --------------
    # plot - prepare

    if dax is None:

        if fs is None:
            fs = (14, 8)

        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.95,
                'bottom': 0.05, 'top': 0.90,
                'hspace': 0.15, 'wspace': 0.3,
            }

        fig = plt.figure(figsize=fs)
        fig.suptitle(key, size=14, fontweight='bold')
        gs = gridspec.GridSpec(nrows=3, ncols=9, **dmargin)

        # n0, n1
        ax0 = fig.add_subplot(gs[:2, :2], aspect='auto')
        ax0.set_ylabel(lab00)
        ax0.set_xlabel(lab01)
        ax0.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax0.xaxis.set_label_position('top')

        ax1 = fig.add_subplot(gs[:2, 2], sharey=ax0)
        ax1.set_xlabel('data')
        ax1.set_ylabel(lab00)
        ax1.tick_params(
            axis="y",
            left=False, right=True,
            labelleft=False, labelright=True,
        )
        ax1.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax1.yaxis.set_label_position('right')
        ax1.xaxis.set_label_position('top')

        ax2 = fig.add_subplot(gs[2, :2], sharex=ax0)
        ax2.set_ylabel('data')
        ax2.set_xlabel(lab01)

        ax1.set_xlim(ymin, ymax)
        ax2.set_ylim(ymin, ymax)

        # n0, n2
        ax3 = fig.add_subplot(gs[:2, 3:5], aspect='auto')
        ax3.set_ylabel(lab10)
        ax3.set_xlabel(lab11)
        ax3.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax3.xaxis.set_label_position('top')

        ax4 = fig.add_subplot(gs[2, 3:5], sharey=ax3)
        ax4.set_xlabel('data')
        ax4.set_ylabel(lab10)
        ax4.tick_params(
            axis="y",
            left=False, right=True,
            labelleft=False, labelright=True,
        )
        ax4.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax4.yaxis.set_label_position('right')
        ax4.xaxis.set_label_position('top')

        ax5 = fig.add_subplot(gs[:2, 5], sharex=ax3)
        ax5.set_ylabel('data')
        ax5.set_xlabel(lab11)

        ax4.set_xlim(ymin, ymax)
        ax5.set_ylim(ymin, ymax)

        # n1, n2
        ax6 = fig.add_subplot(gs[:2, 6:8], aspect='auto')
        ax6.set_ylabel(lab20)
        ax6.set_xlabel(lab21)
        ax6.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax6.xaxis.set_label_position('top')

        ax7 = fig.add_subplot(gs[2, 6:8], sharey=ax6)
        ax7.set_xlabel('data')
        ax7.set_ylabel(lab20)
        ax7.tick_params(
            axis="y",
            left=False, right=True,
            labelleft=False, labelright=True,
        )
        ax7.tick_params(
            axis="x",
            bottom=False, top=True,
            labelbottom=False, labeltop=True,
        )
        ax7.yaxis.set_label_position('right')
        ax7.xaxis.set_label_position('top')

        ax8 = fig.add_subplot(gs[:2, 8], sharex=ax6)
        ax8.set_ylabel('data')
        ax8.set_xlabel(lab21)

        ax7.set_xlim(ymin, ymax)
        ax8.set_ylim(ymin, ymax)

        dax = {
            'matrix0': {'handle': ax0, 'type': 'matrix', 'inverty': True},
            'matrix1': {'handle': ax3, 'type': 'matrix', 'inverty': True},
            'matrix2': {'handle': ax6, 'type': 'matrix', 'inverty': True},
            'vertical0': {'handle': ax1, 'type': 'misc'},
            'vertical1': {'handle': ax4, 'type': 'misc'},
            'vertical2': {'handle': ax7, 'type': 'misc'},
            'horizontal0': {'handle': ax2, 'type': 'misc'},
            'horizontal1': {'handle': ax5, 'type': 'misc'},
            'horizontal2': {'handle': ax8, 'type': 'misc'},
        }

    dax = _generic_check._check_dax(dax=dax, main='matrix')

    # ---------------
    # plot fixed part

    # ----------------
    # define and set dgroup

    dgroup = {
        'i0': {
            'ref': [ref0],
            'data': ['index'],
            'nmax': nmax,
        },
        'i1': {
            'ref': [ref1],
            'data': ['index'],
            'nmax': nmax,
        },
        'i2': {
            'ref': [ref2],
            'data': ['index'],
            'nmax': nmax,
        },
    }

    # -----------------
    # plot first slice

    kax = 'matrix0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # image => ref2
        im = ax.imshow(
            data[:, :, ind[2]],
            extent=extent0,
            interpolation='nearest',
            origin='upper',
            aspect=aspect,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )
        ax.invert_yaxis()

        kim = 'im0'
        coll.add_mobile(
            key=kim,
            handle=im,
            ref=ref2,
            data='index',
            dtype='data',
            ax=kax,
            ind=0,
        )

        # lh, lv => ref0, ref1
        ii = 0
        lh = ax.axhline(ind[0], c=color_dict['i0'][ii], lw=1., ls='-')
        lv = ax.axvline(ind[1], c=color_dict['i0'][ii], lw=1., ls='-')

        # update coll
        kh = f'lh0-{ii:02.0f}'
        kv = f'lv0-{ii:02.0f}'
        coll.add_mobile(
            key=kh,
            handle=lh,
            ref=ref0,
            data='index',
            dtype='ydata',
            ax=kax,
            ind=ii,
        )
        coll.add_mobile(
            key=kv,
            handle=lv,
            ref=ref1,
            data='index',
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refx=[ref1], refy=[ref0])

    kax = 'vertical0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ii = 0
        l0, = ax.plot(
            data[:, ind[1], ind[2]],
            np.arange(0, n0),
            ls='-',
            marker='.',
            lw=1.,
            color=color_dict['i0'][ii],
            label=f'ind0 = {ind[0]}',
        )

        km = f'vprof0'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref1,),
            data=key,
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        l0 = ax.axhline(
            ind[1],
            c=color_dict['i0'][ii],
        )
        km = f'lh-v0'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref0,),
            data='index',
            dtype='ydata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refy=[ref0])

    kax = 'horizontal0'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ii = 0
        l1, = ax.plot(
            np.arange(0, n1),
            data[ind[0], :, ind[2]],
            ls='-',
            marker='.',
            lw=1.,
            color=color_dict['i0'][ii],
            label=f'ind1 = {ind[1]}',
        )

        km = f'hprof0-h{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l1,
            ref=(ref0,),
            data=key,
            dtype='ydata',
            ax=kax,
            ind=ii,
        )

        l0 = ax.axvline(
            ind[0],
            c=color_dict['i0'][ii],
        )
        km = f'lv0-h{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref1,),
            data='index',
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refx=[ref1])

    # -----------------
    # plot second slice

    kax = 'matrix1'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # image => ref2
        im = ax.imshow(
            data[:, ind[1], :],
            extent=extent1,
            interpolation='nearest',
            origin='upper',
            aspect=aspect,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )
        ax.invert_yaxis()

        kim = 'im1'
        coll.add_mobile(
            key=kim,
            handle=im,
            ref=ref1,
            data='index',
            dtype='data',
            ax=kax,
            ind=0,
        )

        # lh, lv => ref0, ref1
        ii = 0
        lh = ax.axhline(ind[0], c=color_dict['i1'][ii], lw=1., ls='-')
        lv = ax.axvline(ind[2], c=color_dict['i1'][ii], lw=1., ls='-')

        # update coll
        kh = f'lh1-{ii:02.0f}'
        kv = f'lv1-{ii:02.0f}'
        coll.add_mobile(
            key=kh,
            handle=lh,
            ref=ref0,
            data='index',
            dtype='ydata',
            ax=kax,
            ind=ii,
        )
        coll.add_mobile(
            key=kv,
            handle=lv,
            ref=ref2,
            data='index',
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refx=[ref2], refy=[ref0])

    kax = 'vertical1'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ii = 0
        l0, = ax.plot(
            data[:, ind[1], ind[2]],
            np.arange(0, n0),
            ls='-',
            marker='.',
            lw=1.,
            color=color_dict['i0'][ii],
            label=f'ind0 = {ind[0]}',
        )

        km = f'vprof1-{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref2,),
            data=key,
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        l0 = ax.axhline(
            ind[2],
            c=color_dict['i0'][ii],
        )
        km = f'lh1-v{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref0,),
            data='index',
            dtype='ydata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refy=[ref0])

    kax = 'horizontal1'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ii = 0
        l1, = ax.plot(
            np.arange(0, n2),
            data[ind[0], ind[1], :],
            ls='-',
            marker='.',
            lw=1.,
            color=color_dict['i0'][ii],
            label=f'ind1 = {ind[2]}',
        )

        km = f'hprof1-{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l1,
            ref=(ref0,),
            data=key,
            dtype='ydata',
            ax=kax,
            ind=ii,
        )

        l0 = ax.axvline(
            ind[0],
            c=color_dict['i0'][ii],
        )
        km = f'lv1-h{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref2,),
            data='index',
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refx=[ref2])

    # -----------------
    # plot third slice

    kax = 'matrix2'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # image => ref2
        im = ax.imshow(
            data[ind[0], :, :],
            extent=extent1,
            interpolation='nearest',
            origin='upper',
            aspect=aspect,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )
        ax.invert_yaxis()

        kim = 'im2'
        coll.add_mobile(
            key=kim,
            handle=im,
            ref=ref1,
            data='index',
            dtype='data',
            ax=kax,
            ind=0,
        )

        # lh, lv => ref0, ref1
        ii = 0
        lh = ax.axhline(ind[1], c=color_dict['i1'][ii], lw=1., ls='-')
        lv = ax.axvline(ind[2], c=color_dict['i1'][ii], lw=1., ls='-')

        # update coll
        kh = f'lh2-{ii:02.0f}'
        kv = f'lv2-{ii:02.0f}'
        coll.add_mobile(
            key=kh,
            handle=lh,
            ref=ref1,
            data='index',
            dtype='ydata',
            ax=kax,
            ind=ii,
        )
        coll.add_mobile(
            key=kv,
            handle=lv,
            ref=ref2,
            data='index',
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refx=[ref2], refy=[ref1])

    kax = 'vertical2'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ii = 0
        l0, = ax.plot(
            data[ind[0], :, ind[2]],
            np.arange(0, n1),
            ls='-',
            marker='.',
            lw=1.,
            color=color_dict['i0'][ii],
            label=f'ind0 = {ind[0]}',
        )

        km = f'vprof2-{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref2,),
            data=key,
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        l0 = ax.axhline(
            ind[2],
            c=color_dict['i0'][ii],
        )
        km = f'lh2-v{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref1,),
            data='index',
            dtype='ydata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refy=[ref0])

    kax = 'horizontal2'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ii = 0
        l1, = ax.plot(
            np.arange(0, n2),
            data[ind[0], ind[1], :],
            ls='-',
            marker='.',
            lw=1.,
            color=color_dict['i0'][ii],
            label=f'ind1 = {ind[2]}',
        )

        km = f'hprof2-{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l1,
            ref=(ref1,),
            data=key,
            dtype='ydata',
            ax=kax,
            ind=ii,
        )

        l0 = ax.axvline(
            ind[1],
            c=color_dict['i0'][ii],
        )
        km = f'lv2-h{ii:02.0f}'
        coll.add_mobile(
            key=km,
            handle=l0,
            ref=(ref2,),
            data='index',
            dtype='xdata',
            ax=kax,
            ind=ii,
        )

        dax[kax].update(refx=[ref2])

    # add axes
    for kax in dax.keys():
        coll.add_axes(key=kax, **dax[kax])

    # increment dict

    coll.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)

    # connect
    if connect is True:
        coll.connect()

    return coll



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


import os


import numpy as np
import matplotlib.pyplot as plt


_PATH_HERE = os.path.dirname(__file__)
_PATH_TOFU = os.path.dirname(os.path.dirname(_PATH_HERE))


import tofu as tf
import tofu.geom as tfg
import tofu.tests.tests01_geom.test_05_solid_angles as tftests


# #############################################################################
# #############################################################################
#               Benchmark of Mesh2D routines
#                   High level routines
# #############################################################################


class SolidAngle:
    """ Benchmark the solid angle computation routines

    In particular:

    """

    # -----------------------------
    # Attributes reckognized by asv

    # time before benchmark is killed
    timeout = 500
    repeat = (1, 10, 20.0)
    sample_time = 0.100

    # -------------------------------------------------------
    # Setup and teardown, run before / after benchmark methods

    def setup_cache(self):
        return 0

    def setup(self, out):
        """ run before each benchmark method, out from setup_cache  """
        self.light = tftests._create_light(npts=10000)
        self.visibility = tftests._create_visibility(npts=4)
        self.etendue = tftests._create_etendue(res=np.r_[0.001])

    def teardown(self, out):
        pass

    # -------------------------------------
    # benchmarks methods for solid angle

    def time_00_light(self, out):
        sa = tfg.calc_solidangle_apertures(
            pts_x=self.light['pts_x'],
            pts_y=self.light['pts_y'],
            pts_z=self.light['pts_z'],
            apertures=self.light['ap']['ap3'],
            detectors=self.light['det'],
            visibility=False,
            return_vector=False,
        )

    def time_01_unitvectors(self, out):
        sa, uv_x, uv_y, uv_z = tfg.calc_solidangle_apertures(
            pts_x=self.light['pts_x'],
            pts_y=self.light['pts_y'],
            pts_z=self.light['pts_z'],
            apertures=self.light['ap']['ap3'],
            detectors=self.light['det'],
            visibility=False,
            return_vector=True,
        )

    # def time_02_visibility(self, out):
        # sa = tfg.calc_solidangle_apertures(
            # pts_x=self.visibility['pts_x'],
            # pts_y=self.visibility['pts_y'],
            # pts_z=self.visibility['pts_z'],
            # apertures=self.visibility['ap'],
            # detectors=self.visibility['det'],
            # config=self.visibility['config'],
            # visibility=True,
            # return_vector=False,
        # )

    def time_03_etendue_check(self, out):
        detend1 = tfg.compute_etendue(
            det=self.etendue['det'],
            aperture=self.etendue['aperture'],
            res=self.etendue['res'],
            check=True,
            verb=False,
            analytical=False,
            plot=False,
        )

    def time_04_etendue_summed(self, out):
        detend1 = tfg.compute_etendue(
            det=self.etendue['det'],
            aperture=self.etendue['aperture'],
            res=self.etendue['res'],
            check=False,
            verb=False,
            analytical=False,
            plot=False,
        )
